#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
The TrueType font format was developed by Apple Computer, Inc., and has been adopted as a standard font
format for the Microsoft Windows operating system. Specifications for the TrueType font file format are
available in Apple’s TrueType Reference Manual and Microsoft’s TrueType 1.0 Font Files Technical
Specification (see Bibliography).
"""
import io
import typing
import zlib
from decimal import Decimal
from pathlib import Path

from fontTools.agl import toUnicode  # type: ignore [import]
from fontTools.pens.boundsPen import BoundsPen  # type: ignore [import]
from fontTools.ttLib import TTFont  # type: ignore [import]

from borb.io.read.types import Decimal as bDecimal
from borb.io.read.types import Dictionary, List, Name, Stream, String
from borb.pdf.canvas.font.composite_font.cid_font_type_2 import CIDType2Font
from borb.pdf.canvas.font.composite_font.font_type_0 import Type0Font
from borb.pdf.canvas.font.simple_font.font_type_1 import Type1Font


class TrueTypeFont(Type1Font):
    """
    A TrueType font dictionary may contain the same entries as a Type 1 font dictionary (see Table 111), with these
    differences:
    •   The value of Subtype shall be TrueType.
    •   The value of Encoding is subject to limitations that are described in 9.6.6, "Character Encoding".
    •   The value of BaseFont is derived differently. The PostScript name for the value of BaseFont may be determined in one of two ways:
    •   If the TrueType font program's “name” table contains a PostScript name, it shall be used.
    •   In the absence of such an entry in the “name” table, a PostScript name shall be derived from the name by
        which the font is known in the host operating system. On a Windows system, the name shall be based on
        the lfFaceName field in a LOGFONT structure; in the Mac OS, it shall be based on the name of the FOND
        resource. If the name contains any SPACEs, the SPACEs shall be removed.
    """

    @staticmethod
    def true_type_font_from_file(
        font_file: typing.Union[Path, bytes],
    ) -> typing.Union["TrueTypeFont", "Type0Font"]:
        """
        This function returns the PDF TrueTypeFont object for a given TTF file
        """
        if isinstance(font_file, Path):
            assert font_file.exists()
            assert font_file.name.endswith(".ttf")

        # read bytes if needed
        font_file_bytes: typing.Optional[bytes] = None
        if isinstance(font_file, Path):
            with open(font_file, "rb") as ffh:
                font_file_bytes = ffh.read()
            assert font_file_bytes
        else:
            font_file_bytes = font_file

        # read file
        ttf_font_file: TTFont = TTFont(io.BytesIO(font_file_bytes))

        # read cmap
        cmap: typing.Optional[typing.Dict[int, str]] = ttf_font_file.getBestCmap()
        assert cmap is not None
        cmap_reverse: typing.Dict[str, int] = {}
        for k, v in cmap.items():
            if v in cmap_reverse:
                cmap_reverse[v] = min(cmap_reverse[v], k)
            else:
                cmap_reverse[v] = k
        glyph_order: typing.List[str] = [
            x for x in ttf_font_file.glyphOrder if x in cmap_reverse
        ]

        # if there are more than 256 glyphs, we need to switch to a Type0Font
        if len(glyph_order) >= 256:
            # fmt: off
            type_0_font: Type0Font = TrueTypeFont._type_0_font_from_file(ttf_font_file)
            type_0_font["DescendantFonts"][0]["FontDescriptor"][Name("FontFile2")] = TrueTypeFont._get_font_file_stream(font_file_bytes)
            return type_0_font
            # fmt: on

        # build font
        font: TrueTypeFont = TrueTypeFont()
        font_name: str = TrueTypeFont._get_base_font(ttf_font_file)
        font[Name("Name")] = Name(font_name)
        font[Name("BaseFont")] = Name(font_name)

        # build widths
        units_per_em: bDecimal = bDecimal(ttf_font_file["head"].unitsPerEm)
        if cmap is not None:
            # fmt: off
            font[Name("FirstChar")] = bDecimal(min([cid for cid, _ in enumerate(glyph_order)]))
            font[Name("LastChar")] = bDecimal(max([cid for cid, _ in enumerate(glyph_order)]))
            # fmt: on
            font[Name("Widths")] = List()
            for glyph_name in glyph_order:
                w: typing.Union[bDecimal, Decimal] = (
                    bDecimal(ttf_font_file.getGlyphSet()[glyph_name].width)
                    / units_per_em
                ) * Decimal(1000)
                w = bDecimal(round(w, 2))
                font["Widths"].append(w)

        assert font[Name("FirstChar")] >= 0
        assert (
            font[Name("LastChar")] < 256
        ), "TrueType fonts with more than 256 glyphs are currently not supported."

        font[Name("FontDescriptor")] = TrueTypeFont._get_font_descriptor(ttf_font_file)
        font[Name("Encoding")] = Dictionary()
        font["Encoding"][Name("BaseEncoding")] = Name("WinAnsiEncoding")
        font["Encoding"][Name("Differences")] = List()
        for i in range(0, len(glyph_order)):
            font["Encoding"]["Differences"].append(bDecimal(i))
            font["Encoding"]["Differences"].append(Name(glyph_order[i]))

        # embed font file
        # fmt: off
        font["FontDescriptor"][Name("FontFile2")] = TrueTypeFont._get_font_file_stream(font_file_bytes)
        # fmt: on

        # return
        return font

    @staticmethod
    def _get_font_file_stream(font_file_bytes: bytes) -> Stream:
        font_stream: Stream = Stream()
        font_stream[Name("Type")] = Name("Font")
        font_stream[Name("Length")] = bDecimal(len(font_file_bytes))
        font_stream[Name("Length1")] = bDecimal(len(font_file_bytes))
        font_stream[Name("Filter")] = Name("FlateDecode")
        font_stream[Name("DecodedBytes")] = font_file_bytes
        font_stream[Name("Bytes")] = zlib.compress(font_file_bytes, 9)
        return font_stream

    @staticmethod
    def _get_font_descriptor(ttf_font_file: TTFont) -> Dictionary:

        # fmt: off
        font_descriptor: Dictionary = Dictionary()
        font_descriptor[Name("Type")] = Name("FontDescriptor")
        font_descriptor[Name("FontName")] = String(TrueTypeFont._get_base_font(ttf_font_file))
        font_descriptor[Name("FontStretch")] = Name("Normal")  # TODO
        font_descriptor[Name("FontWeight")] = bDecimal(400)  # TODO
        font_descriptor[Name("Flags")] = bDecimal(4)  # TODO
        # fmt: on

        # determine FontBBox, CapHeight
        units_per_em: float = ttf_font_file["head"].unitsPerEm
        min_x: float = 1000
        min_y: float = 1000
        max_x: float = 0
        max_y: float = 0
        cap_height: typing.Optional[bDecimal] = None
        glyph_set = ttf_font_file.getGlyphSet()
        for glyph_name in ttf_font_file.glyphOrder:
            pen = BoundsPen(glyph_set)
            glyph_set[glyph_name].draw(pen)
            if pen.bounds is None:
                continue
            # determine CapHeight
            if glyph_name in "EFHIJLMNTZ" and cap_height is None:
                cap_height = bDecimal(pen.bounds[3])
            min_x = min(min_x, pen.bounds[0] / units_per_em * 1000)
            min_y = min(min_y, pen.bounds[1] / units_per_em * 1000)
            max_x = max(max_x, pen.bounds[2] / units_per_em * 1000)
            max_y = max(max_y, pen.bounds[3] / units_per_em * 1000)
        if cap_height is None:
            cap_height = bDecimal(840)

        font_descriptor[Name("FontBBox")] = List().set_is_inline(True)  # type: ignore[attr-defined]
        font_descriptor["FontBBox"].append(bDecimal(min_x))
        font_descriptor["FontBBox"].append(bDecimal(min_y))
        font_descriptor["FontBBox"].append(bDecimal(max_x))
        font_descriptor["FontBBox"].append(bDecimal(max_y))

        # fmt: off
        font_descriptor[Name("ItalicAngle")] = bDecimal(ttf_font_file["post"].italicAngle)
        font_descriptor[Name("Ascent")] = bDecimal(ttf_font_file["hhea"].ascent / units_per_em * 1000)
        font_descriptor[Name("Descent")] = bDecimal(ttf_font_file["hhea"].descent / units_per_em * 1000)
        font_descriptor[Name("CapHeight")] = cap_height
        font_descriptor[Name("StemV")] = bDecimal(297)             # TODO
        # fmt: on

        return font_descriptor

    @staticmethod
    def _get_base_font(ttf_font_file: TTFont) -> str:
        font_name: str = str(
            [
                x
                for x in ttf_font_file["name"].names
                if x.platformID == 3 and x.platEncID == 1 and x.nameID == 6
            ][0].string,
            "latin1",
        )
        font_name = "".join(
            [x for x in font_name if x.lower() in "abcdefghijklmnopqrstuvwxyz-"]
        )
        return font_name

    @staticmethod
    def _build_custom_cmap_for_type_0_font(ttf_font_file: TTFont) -> Stream:
        cmap_prefix: str = """
        /CIDInit /ProcSet findresource begin
        12 dict begin
        begincmap
        /CIDSystemInfo <</Registry (Adobe) /Ordering (UCS) /Supplement 0>> def
        /CMapName /Adobe-Identity-UCS def
        /CMapType 2 def
        1 begincodespacerange
        <0000> <FFFF>
        endcodespacerange
        """
        # 1 beginbfchar
        # <0000> <0000>
        # endbfchar
        pairs: typing.List[typing.Tuple[str, str]] = []
        for cid, g in enumerate(ttf_font_file.glyphOrder):
            g_unicode: str = toUnicode(g)
            if len(g_unicode) == 0:
                continue
            g_hex: str = ""
            if len(g_unicode) == 1:
                g_hex = hex(ord(g_unicode))[2:]
            if len(g_unicode) == 2:
                g_hex = hex(ord(g_unicode[0]))[2:] + hex(ord(g_unicode[1]))[2:]
            while len(g_hex) < 4:
                g_hex = "0" + g_hex
            i_hex: str = hex(cid)[2:]
            while len(i_hex) < 4:
                i_hex = "0" + i_hex
            pairs.append((i_hex, g_hex))

        # split in lots of 100
        cmap_content: str = ""
        for i in range(0, len(pairs), 100):
            start_index: int = i
            end_index: int = min(start_index + 100, len(pairs))
            n: int = end_index - start_index
            cmap_content += "%d beginbfchar\n" % n
            for j in range(start_index, end_index):
                cmap_content += "<%s> <%s>\n" % (pairs[j][0], pairs[j][1])
            cmap_content += "endbfchar\n"

        cmap_suffix: str = (
            "endcmap\nCMapName currentdict /CMap defineresource pop\nend\nend\n"
        )

        bts: bytes = (cmap_prefix + cmap_content + cmap_suffix).encode("latin1")
        to_unicode_stream = Stream()
        to_unicode_stream[Name("DecodedBytes")] = bts
        to_unicode_stream[Name("Bytes")] = zlib.compress(bts, 9)
        to_unicode_stream[Name("Filter")] = Name("FlateDecode")
        to_unicode_stream[Name("Length")] = bDecimal(len(bts))
        return to_unicode_stream

    @staticmethod
    def _build_custom_widths_array_for_type_0_font(ttf_font_file: TTFont) -> List:
        units_per_em: bDecimal = bDecimal(ttf_font_file["head"].unitsPerEm)
        cmap = ttf_font_file.getBestCmap()
        glyph_set = ttf_font_file.getGlyphSet()
        widths_array: List = List()
        for cid, g in enumerate(ttf_font_file.glyphOrder):
            glyph_width: bDecimal = bDecimal(0)
            try:
                glyph_width = bDecimal(
                    glyph_set[cmap[ord(toUnicode(g))]].width / units_per_em * 1000
                )
            except:
                pass
            widths_array.append(bDecimal(cid))
            widths_array.append(List())
            widths_array[-1].append(bDecimal(glyph_width))
        return widths_array

    @staticmethod
    def _type_0_font_from_file(ttf_font_file: TTFont) -> "Type0Font":
        type_0_font: Type0Font = Type0Font()

        # build BaseFont
        font_name: str = TrueTypeFont._get_base_font(ttf_font_file)
        type_0_font[Name("BaseFont")] = Name(font_name)

        # set Encoding
        type_0_font[Name("Encoding")] = Name("Identity-H")

        # set ToUnicode
        type_0_font[
            Name("ToUnicode")
        ] = TrueTypeFont._build_custom_cmap_for_type_0_font(ttf_font_file)

        # build DescendantFont
        descendant_font: CIDType2Font = CIDType2Font()
        descendant_font[Name("Type")] = Name("Font")
        descendant_font[Name("Subtype")] = Name("CIDFontType2")
        descendant_font[Name("BaseFont")] = Name(font_name)
        descendant_font[Name("FontDescriptor")] = TrueTypeFont._get_font_descriptor(
            ttf_font_file
        )
        descendant_font[Name("DW")] = bDecimal(250)

        # build W array
        descendant_font[
            Name("W")
        ] = TrueTypeFont._build_custom_widths_array_for_type_0_font(ttf_font_file)
        descendant_font[Name("CIDToGIDMap")] = Name("Identity")

        # build CIDSystemInfo
        # fmt: off
        descendant_font[Name("CIDSystemInfo")] = Dictionary()
        descendant_font[Name("CIDSystemInfo")][Name("Registry")] = String("Adobe")
        descendant_font[Name("CIDSystemInfo")][Name("Ordering")] = String("Identity")
        descendant_font[Name("CIDSystemInfo")][Name("Supplement")] = bDecimal(0)
        # fmt: on

        # add to DescendantFonts
        type_0_font[Name("DescendantFonts")] = List()
        type_0_font[Name("DescendantFonts")].append(descendant_font)

        # return
        return type_0_font

    def __init__(self):
        super(TrueTypeFont, self).__init__()
        self[Name("Subtype")] = Name("TrueType")

    def _empty_copy(self) -> "Font":  # type: ignore [name-defined]
        return TrueTypeFont()

    def __deepcopy__(self, memodict={}):
        # fmt: off
        f_out: TrueTypeFont = super(TrueTypeFont, self).__deepcopy__(memodict)
        f_out[Name("Subtype")] = Name("TrueType")
        f_out._character_identifier_to_unicode_lookup: typing.Dict[int, str] = {k: v for k, v in self._character_identifier_to_unicode_lookup.items()}
        f_out._unicode_lookup_to_character_identifier: typing.Dict[str, int] = {k: v for k, v in self._unicode_lookup_to_character_identifier.items()}
        return f_out
        # fmt: on



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This implementation of EventListener exports a Page as an mp3 file, essentially reading the text on the Page
"""
import tempfile
import typing
from decimal import Decimal
from pathlib import Path

from gtts import gTTS  # type: ignore [import]

from borb.pdf.canvas.geometry.rectangle import Rectangle
from borb.pdf.canvas.layout.text.paragraph import Paragraph
from borb.pdf.page.page import Page
from borb.pdf.pdf import PDF
from borb.toolkit.text.simple_paragraph_extraction import SimpleParagraphExtraction


class PDFToMP3(SimpleParagraphExtraction):
    """
    This implementation of EventListener exports a Page as an mp3 file, essentially reading the text on the Page
    """

    @staticmethod
    def convert_pdf_to_mp3(file: Path, page_number: int) -> Path:
        """
        This function converts a PDF to an MP3 file, returning its Path
        """
        l: "PDFToMP3" = PDFToMP3()
        with open(file, "rb") as pdf_file_handle:
            PDF.loads(pdf_file_handle, [l])  # type: ignore [arg-type]
        temporary_file: Path = Path(
            tempfile.NamedTemporaryFile(prefix="pdf_to_mp3", suffix=".mp3").name
        )
        l.get_audio_for_page(page_number, str(temporary_file))
        return temporary_file

    def __init__(
        self,
        include_position: bool = True,
        language: str = "en",
        slow: bool = False,
    ):
        """
        Constructs a new PDFToMP3
        """
        super(PDFToMP3, self).__init__()

        # tts info
        self._include_position = include_position
        self._language = language
        self._speak_slowly = slow

        # page info
        self._text_to_speak_for_page: typing.Dict[int, str] = {}

    def _end_page(self, page: Page):
        super(PDFToMP3, self)._end_page(page)
        self._text_to_speak_for_page[self._current_page_number] = "".join(
            [
                self._get_text_for_paragraph(p, i, self._current_page_number)
                for i, p in enumerate(
                    self._paragraphs_per_page[self._current_page_number]
                )
            ]
        )

    def _get_text_for_x(self, bounding_box: Rectangle) -> str:
        assert self._current_page
        w = self._current_page.get_page_info().get_width()
        assert w is not None
        xs: typing.List[Decimal] = [Decimal(0), w * Decimal(0.33), w * Decimal(0.66), w]
        x = bounding_box.x + bounding_box.width * Decimal(0.5)
        if xs[0] <= x <= xs[1]:
            return "left"
        if xs[1] <= x <= xs[2]:
            return "center"
        if xs[2] <= x <= xs[3]:
            return "right"
        return ""

    def _get_text_for_y(self, bounding_box: Rectangle) -> str:
        assert self._current_page
        h = self._current_page.get_page_info().get_height()
        assert h is not None
        assert h is not None
        ys: typing.List[Decimal] = [h, h * Decimal(0.66), h * Decimal(0.33), Decimal(0)]
        y = bounding_box.y
        if ys[1] <= y <= ys[0]:
            return "top"
        if ys[2] <= y <= ys[1]:
            return "middle"
        if ys[3] <= y <= ys[2]:
            return "bottom"
        return ""

    def _get_text_for_paragraph(
        self, paragraph: Paragraph, paragraph_number: int, page_number: int
    ):

        # text to speak
        text_to_speak_for_paragraph = ""

        # position
        if self._include_position:
            lbox: typing.Optional[Rectangle] = paragraph.get_previous_layout_box()
            assert lbox is not None
            text_to_speak_for_paragraph += "Page %d, paragraph %d, %s %s." % (
                page_number + 1,
                paragraph_number + 1,
                self._get_text_for_y(lbox),
                self._get_text_for_x(lbox),
            )
        # text of paragraph
        text_to_speak_for_paragraph += paragraph._text

        # force period if needed
        if text_to_speak_for_paragraph[-1] not in ["?", "!", "."]:
            text_to_speak_for_paragraph += ". "

        # return
        return text_to_speak_for_paragraph

    def get_audio_for_page(self, page_number: int, path: str):
        """
        This function creates and then returns the audio-file for the text spoken at the given page
        """
        sound = gTTS(
            text=self._text_to_speak_for_page[page_number], lang=self._language
        )
        sound.save(path)
        return path



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
    This class represents a single page in a PDF document
"""
import io
import typing
import zlib
from decimal import Decimal

from borb.io.read.types import Decimal as bDecimal
from borb.io.read.types import Dictionary, List, Name, Stream, String
from borb.pdf.canvas.canvas import Canvas
from borb.pdf.canvas.geometry.rectangle import Rectangle
from borb.pdf.canvas.layout.annotation.annotation import Annotation
from borb.pdf.page.page_info import PageInfo


class Page(Dictionary):
    """
    This class represents a single page in a PDF document
    """

    def __init__(self, width: Decimal = Decimal(595), height: Decimal = Decimal(842)):
        super(Page, self).__init__()

        # type
        self[Name("Type")] = Name("Page")

        # size: A4 portrait
        self[Name("MediaBox")] = List().set_is_inline(True)  # type: ignore [attr-defined]
        self["MediaBox"].append(bDecimal(0))
        self["MediaBox"].append(bDecimal(0))
        self["MediaBox"].append(bDecimal(width))
        self["MediaBox"].append(bDecimal(height))

    def get_page_info(self) -> PageInfo:
        """
        This function returns the PageInfo object for this Page
        """
        return PageInfo(self)

    def get_document(self) -> "Document":  # type: ignore [name-defined]
        """
        This function returns the Document from which this Page came
        """
        return self.get_root()  # type: ignore [attr-defined]

    #
    # FORMS
    #

    def has_acroforms(self) -> bool:
        """
        This function returns True if this Page contains fields from an AcroForm
        :return:    True if this Page contains fields from an AcroForm, False otherwise
        """
        return (
            len(
                [
                    x
                    for x in self.get("Annots", [])
                    if "Type" in x
                    and x["Type"] == "Annot"
                    and "Subtype" in x
                    and x["Subtype"] == "Widget"
                    and "FT" in x
                    and x["FT"] in ["Btn", "Ch", "Tx"]
                ]
            )
            != 0
        )

    def has_form_field(self, field_name: str) -> bool:
        """
        This function returns True if this Page contains a form field with the given name
        :param field_name:  the field_name to be queried
        :return:            True if this Page contains a form field with the given field_name
        """
        assert len(field_name) != 0
        return (
            len(
                [
                    x
                    for x in self.get("Annots", [])
                    if "Type" in x
                    and x["Type"] == "Annot"
                    and "Subtype" in x
                    and x["Subtype"] == "Widget"
                    and "FT" in x
                    and x["FT"] in ["Btn", "Ch", "Tx"]
                    and "T" in x
                    and x["T"] == field_name
                ]
            )
            != 0
        )

    def get_form_field_value(
        self, field_name: str
    ) -> typing.Optional[typing.Union[str, bool]]:
        """
        This function returns the value of the form field with the given field_name
        :param field_name:  the field_name of the field to be queried
        :return:            the value of the form field being queried
        """
        field_dictionaries: typing.List[Dictionary] = [
            x
            for x in self.get("Annots", [])
            if "Type" in x
            and x["Type"] == "Annot"
            and "Subtype" in x
            and x["Subtype"] == "Widget"
            and "FT" in x
            and x["FT"] in ["Btn", "Ch", "Tx"]
            and "T" in x
            and x["T"] == field_name
        ]
        assert len(field_dictionaries) == 1
        assert "V" in field_dictionaries[0]
        return field_dictionaries[0]["V"]

    def set_form_field_value(self, field_name: str, value: str) -> "Page":
        """
        This function sets the value of the form field with the given field_name
        This function returns self
        :param field_name:  the field_name of the field being queried
        :param value:       the new value of the field
        :return:            self
        """
        field_dictionaries: typing.List[Dictionary] = [
            x
            for x in self.get("Annots", [])
            if "Type" in x
            and x["Type"] == "Annot"
            and "Subtype" in x
            and x["Subtype"] == "Widget"
            and "FT" in x
            and x["FT"] in ["Btn", "Ch", "Tx"]
            and "T" in x
            and x["T"] == field_name
        ]
        assert len(field_dictionaries) == 1
        assert "V" in field_dictionaries[0]
        field_dictionaries[0][Name("V")] = String(value)
        return self

    #
    # ROTATE
    #

    def rotate_right(self) -> "Page":
        """
        This function rotates the Page clockwise by 90 degrees.
        This function returns self.
        """
        # get current rotation
        angle: int = 0
        if "Rotate" in self:
            angle = int(self["Rotate"])

        # rotate left
        angle = (angle + 90) % 360

        # write entry
        if angle == 0 and "Rotate" in self:
            self.pop("Rotate")
        else:
            self[Name("Rotate")] = bDecimal(angle)

        # return
        return self

    def rotate_left(self) -> "Page":
        """
        This function rotates the Page counterclockwise by 90 degrees.
        This function returns self.
        """

        # get current rotation
        angle: int = 0
        if "Rotate" in self:
            angle = int(self["Rotate"])

        # rotate left
        angle = (angle + 270) % 360

        # write entry
        if angle == 0 and "Rotate" in self:
            self.pop("Rotate")
        else:
            self[Name("Rotate")] = bDecimal(angle)

        # return
        return self

    #
    # ANNOTATIONS
    #

    def get_annotations(self) -> List:
        """
        This function returns the annotation(s) on this Page
        """
        if "Annots" not in self:
            self[Name("Annots")] = List()
        return self["Annots"]

    def add_annotation(self, annotation: Annotation) -> "Page":
        """
        This function appends an Annotation to this Page, returning self.
        :param annotation:  the Annotation object to append to this Page
        :return:            self
        """

        # (Optional; PDF 1.4) The annotation name, a text string uniquely
        # identifying it among all the annotations on its page.
        len_annots = len(self["Annots"]) if "Annots" in self else 0
        annotation[Name("NM")] = String("annotation-{0:03d}".format(len_annots))

        # (Optional except as noted below; PDF 1.3; not used in FDF files) An
        # indirect reference to the page object with which this annotation is
        # associated.
        # This entry shall be present in screen annotations associated with
        # rendition actions (PDF 1.5; see 12.5.6.18, “Screen Annotations” and
        # 12.6.4.13, “Rendition Actions”).
        annotation[Name("P")] = self

        # append to /Annots
        if "Annots" not in self:
            self[Name("Annots")] = List()
            self["Annots"].set_parent(self)
        assert isinstance(self["Annots"], List)
        self["Annots"].append(annotation)

        # FreeTextAnnotation needs to embed resources in the Page
        if "Subtype" in annotation and annotation["Subtype"] == "FreeText":
            annotation._embed_font_in_page(self)  # type: ignore [attr-defined]

        # return
        return self

    def apply_redact_annotations(self):
        """
        This function applies the redaction annotations on this Page
        """
        from borb.pdf.canvas.redacted_canvas_stream_processor import (
            RedactedCanvasStreamProcessor,
        )

        rectangles_to_redact: typing.List[Rectangle] = [
            Rectangle(
                x["Rect"][0],
                x["Rect"][1],
                x["Rect"][2] - x["Rect"][0],
                x["Rect"][3] - x["Rect"][1],
            )
            for x in self["Annots"]
            if "Subtype" in x and x["Subtype"] == "Redact" and "Rect" in x
        ]

        # apply redaction
        redacted_canvas_content: bytes = (
            RedactedCanvasStreamProcessor(self, Canvas(), rectangles_to_redact)
            .read(io.BytesIO(self["Contents"]["DecodedBytes"]), [])
            .get_redacted_content()  # type: ignore [attr-defined]
        )

        # update Page Contents (Stream)
        self["Contents"][Name("DecodedBytes")] = redacted_canvas_content
        self["Contents"][Name("Bytes")] = zlib.compress(
            self["Contents"]["DecodedBytes"], 9
        )
        self["Contents"][Name("Length")] = bDecimal(len(self["Contents"]["Bytes"]))

    def _initialize_page_content_stream(self) -> "Page":  # type: ignore[name-defined]

        # build content stream object
        if "Contents" not in self:
            content_stream = Stream()
            content_stream[Name("DecodedBytes")] = b""
            content_stream[Name("Bytes")] = zlib.compress(
                content_stream["DecodedBytes"], 9
            )
            content_stream[Name("Filter")] = Name("FlateDecode")
            content_stream[Name("Length")] = bDecimal(len(content_stream["Bytes"]))

            # set content of page
            self[Name("Contents")] = content_stream

        # set Resources
        if "Resources" not in self:
            self[Name("Resources")] = Dictionary()

        # return
        return self

    def _append_to_content_stream(self, s: str) -> "Page":  # type: ignore[name-defined]
        self._initialize_page_content_stream()
        content_stream = self["Contents"]

        # prepend whitespace if needed
        if len(content_stream[Name("DecodedBytes")]) != 0:
            # fmt: off
            decoded_bytes_last_char: str = str(content_stream["DecodedBytes"][-1:], encoding="latin1")
            if decoded_bytes_last_char not in [" ", "\t", "\n"] and s[0] not in [" ", "\t", "\n"]:
                instructions = " " + s
            # fmt: on

        content_stream[Name("DecodedBytes")] += s.encode("latin1")
        content_stream[Name("Bytes")] = zlib.compress(content_stream["DecodedBytes"], 9)
        content_stream[Name("Length")] = bDecimal(len(content_stream["Bytes"]))

        # return
        return self



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This class contains utility methods for using the Table classes in borb.
"""
import numbers
import typing

from borb.io.read.types import Decimal
from borb.pdf.canvas.color.color import HexColor
from borb.pdf.canvas.layout.table.fixed_column_width_table import FixedColumnWidthTable
from borb.pdf.canvas.layout.table.flexible_column_width_table import (
    FlexibleColumnWidthTable,
)
from borb.pdf.canvas.layout.table.table import TableCell, Table
from borb.pdf.canvas.layout.text.paragraph import Paragraph


class TableUtil:
    """
    This class contains utility methods for using the Table classes in borb.
    """

    @staticmethod
    def from_2d_array(
        data: typing.List[typing.List[typing.Any]],
        header_row: bool = True,
        header_col: bool = False,
        round_to_n_digits: typing.Optional[int] = None,
        flexible_column_width: bool = True,
    ) -> Table:
        """
        This function creates a Table from a 2D array of (stringable) data
        :param data:                    the data used to populate the Table
        :param header_row:              whether there is a header row
        :param header_col:              whether there is a header column
        :param round_to_n_digits:       this value is None if digits should not be rounded, if this value is not None, digits are rounded to this precision
        :param flexible_column_width:   true if a FlexibleColumnWidthTable should be used, false otherwise
        :return:                        a Table containing the data
        """

        # get number of rows
        row_count: int = len(data)
        assert row_count > 0, "Table must contain at least 1 row"

        # get number of cols
        col_count: int = len(data[0])
        assert col_count > 0, "Table must contain at least 1 column"
        assert all(
            [len(x) == col_count for x in data]
        ), "All rows must contain the same number of columns"

        # instantiate Table
        t: typing.Optional[Table] = None
        if flexible_column_width:
            t = FlexibleColumnWidthTable(
                number_of_rows=row_count, number_of_columns=col_count
            )
        else:
            t = FixedColumnWidthTable(
                number_of_rows=row_count, number_of_columns=col_count
            )
        assert t is not None

        # add data
        for i in range(0, row_count):
            for j in range(0, col_count):
                # build text
                s: str = ""
                if round_to_n_digits is not None and isinstance(
                    data[i][j], numbers.Number
                ):
                    s = str(round(data[i][j], round_to_n_digits))
                else:
                    s = str(data[i][j])

                # build TableCell
                p: typing.Optional[TableCell] = None
                if (i == 0 and header_row) or (j == 0 and header_col):
                    p = TableCell(
                        Paragraph(s, font_size=Decimal(12), font="Helvetica-Bold"),
                        background_color=HexColor("f1f3f4"),
                    )
                else:
                    p = TableCell(Paragraph(s, font_size=Decimal(12), font="Helvetica"))
                t.add(p)

        # padding
        t.set_padding_on_all_cells(Decimal(3), Decimal(3), Decimal(3), Decimal(3))

        # return
        return t



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This class represents a generic disjoint shape (specified by a List of lines).
It has convenience methods to calculate width and height, perform scaling, etc
"""
import math
import typing
from decimal import Decimal

from borb.pdf.canvas.color.color import HexColor, Color, X11Color
from borb.pdf.canvas.geometry.rectangle import Rectangle
from borb.pdf.canvas.layout.layout_element import Alignment, LayoutElement
from borb.pdf.page.page import Page


class DisconnectedShape(LayoutElement):
    """
    This class represents a generic disjoint shape (specified by a List of lines).
    It has convenience methods to calculate width and height, perform scaling, etc
    """

    def __init__(
        self,
        lines: typing.List[
            typing.Tuple[typing.Tuple[Decimal, Decimal], typing.Tuple[Decimal, Decimal]]
        ],
        stroke_color: Color = HexColor("000000"),
        background_color: typing.Optional[Color] = None,
        border_bottom: bool = False,
        border_color: Color = HexColor("000000"),
        border_left: bool = False,
        border_radius_bottom_left: Decimal = Decimal(0),
        border_radius_bottom_right: Decimal = Decimal(0),
        border_radius_top_left: Decimal = Decimal(0),
        border_radius_top_right: Decimal = Decimal(0),
        border_right: bool = False,
        border_top: bool = False,
        border_width: Decimal = Decimal(1),
        horizontal_alignment: Alignment = Alignment.LEFT,
        line_width: Decimal = Decimal(1),
        margin_bottom: typing.Optional[Decimal] = Decimal(0),
        margin_left: typing.Optional[Decimal] = Decimal(0),
        margin_right: typing.Optional[Decimal] = Decimal(0),
        margin_top: typing.Optional[Decimal] = Decimal(0),
        padding_bottom: Decimal = Decimal(0),
        padding_left: Decimal = Decimal(0),
        padding_right: Decimal = Decimal(0),
        padding_top: Decimal = Decimal(0),
        vertical_alignment: Alignment = Alignment.TOP,
    ):
        super(DisconnectedShape, self).__init__(
            background_color=background_color,
            border_bottom=border_bottom,
            border_color=border_color,
            border_left=border_left,
            border_radius_bottom_left=border_radius_bottom_left,
            border_radius_bottom_right=border_radius_bottom_right,
            border_radius_top_left=border_radius_top_left,
            border_radius_top_right=border_radius_top_right,
            border_right=border_right,
            border_top=border_top,
            border_width=border_width,
            font_size=Decimal(12),
            horizontal_alignment=horizontal_alignment,
            margin_bottom=margin_bottom,
            margin_left=margin_left,
            margin_right=margin_right,
            margin_top=margin_top,
            padding_bottom=padding_bottom,
            padding_left=padding_left,
            padding_right=padding_right,
            padding_top=padding_top,
            vertical_alignment=vertical_alignment,
        )
        assert len(lines) > 0
        self._lines = lines
        self._stroke_color = stroke_color
        self._line_width = line_width

    def get_width(self) -> Decimal:
        """
        This function returns the width of this DisjointShape
        """
        min_x = min([min(x[0][0], x[1][0]) for x in self._lines])
        max_x = max([max(x[0][0], x[1][0]) for x in self._lines])
        return max_x - min_x

    def get_height(self) -> Decimal:
        """
        This function returns the height of this DisjointShape
        """
        min_y = min([min(x[0][1], x[1][1]) for x in self._lines])
        max_y = max([max(x[0][1], x[1][1]) for x in self._lines])
        return max_y - min_y

    def rotate(self, angle_in_radians: float) -> "Shape":  # type: ignore[name-defined]
        """
        This function rotates the DisjointShape for a given angle
        :param angle_in_radians:    the angle
        :return:                    this DisjointShape
        """
        a: Decimal = Decimal(math.cos(angle_in_radians))
        b: Decimal = Decimal(-math.sin(angle_in_radians))
        c: Decimal = Decimal(math.sin(angle_in_radians))
        d: Decimal = Decimal(math.cos(angle_in_radians))
        self._lines = [
            (
                (a * l[0][0] + c * l[0][1], b * l[0][0] + d * l[0][1]),
                (a * l[1][0] + c * l[1][1], b * l[1][0] + d * l[1][1]),
            )
            for l in self._lines
        ]
        return self

    def scale_to_fit(
        self, max_width: Decimal, max_height: Decimal
    ) -> "DisconnectedShape":
        """
        This method scales this DisjointShape to fit a given max. width / height
        """
        w_scale = max_width / self.get_width()
        h_scale = max_height / self.get_height()

        # preserve aspect ratio
        w_scale = min(w_scale, h_scale)
        h_scale = w_scale

        if w_scale < 1:
            self._lines = [
                ((x[0][0] * w_scale, x[0][1]), (x[1][0] * w_scale, x[1][1]))
                for x in self._lines
            ]
        if h_scale < 1:
            self._lines = [
                ((x[0][0], x[0][1] * h_scale), (x[1][0], x[1][1] * h_scale))
                for x in self._lines
            ]
        return self

    def move_to(
        self, lower_left_x: Decimal, lower_left_y: Decimal
    ) -> "DisconnectedShape":
        """
        This method translates this DisjointShape so its lower left corner aligns with the given coordinates
        """
        min_x = min([min(x[0][0], x[1][0]) for x in self._lines])
        min_y = min([min(x[0][1], x[1][1]) for x in self._lines])
        delta_x = lower_left_x - min_x
        delta_y = lower_left_y - min_y
        self._lines = [
            (
                (x[0][0] + delta_x, x[0][1] + delta_y),
                (x[1][0] + delta_x, x[1][1] + delta_y),
            )
            for x in self._lines
        ]
        return self

    def _get_content_box(self, available_space: Rectangle) -> Rectangle:
        return Rectangle(
            available_space.get_x(),
            available_space.get_y() + available_space.get_height() - self.get_height(),
            self.get_width(),
            self.get_height(),
        )

    def _paint_content_box(self, page: Page, available_space: Rectangle) -> None:

        # translate points to fit in box
        self.move_to(
            available_space.get_x(),
            available_space.get_y() + available_space.get_height() - self.get_height(),
        )

        # write content
        stroke_rgb = (self._stroke_color or X11Color("Black")).to_rgb()
        content = "q %f %f %f RG %d w " % (
            float(stroke_rgb.red),
            float(stroke_rgb.green),
            float(stroke_rgb.blue),
            float(self._line_width),
        )
        for l in self._lines:
            content += " %f %f m %f %f l " % (
                float(l[0][0]),
                float(l[0][1]),
                float(l[1][0]),
                float(l[1][1]),
            )

        # stroke
        content += " S Q"

        # append to page
        page._append_to_content_stream(content)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This module contains various implementations of `PageLayout`.
`PageLayout` can be used to add `LayoutElement` objects to a `Page` without
having to specify coordinates.
"""
import copy
import typing
from decimal import Decimal

from borb.pdf.canvas.geometry.rectangle import Rectangle
from borb.pdf.canvas.layout.layout_element import LayoutElement
from borb.pdf.canvas.layout.page_layout.page_layout import PageLayout
from borb.pdf.canvas.layout.text.chunk_of_text import ChunkOfText
from borb.pdf.document.document import Document
from borb.pdf.page.page import Page


class MultiColumnLayout(PageLayout):
    """
    This implementation of PageLayout adds left/right/top/bottom margins to a Page
    and lays out the content on the Page as if there were multiple  columns to flow text, images, etc into.
    Once a column is full, the next column is automatically selected, although the next column can be manually selected.
    """

    def __init__(
        self,
        page: Page,
        number_of_columns: int = 2,
        horizontal_margin: typing.Optional[Decimal] = None,
        vertical_margin: typing.Optional[Decimal] = None,
        inter_column_margin: typing.Optional[Decimal] = None,
    ):
        super().__init__(page)

        # page width/height
        # fmt: off
        self._page_width: typing.Optional[Decimal] = self._page.get_page_info().get_width()
        self._page_height: typing.Optional[Decimal] = self._page.get_page_info().get_height()
        assert self._page_width, "page.get_width() must be defined for MultiColumnLayout to work"
        assert self._page_height, "page.get_height() must be defined for MultiColumnLayout to work"
        # fmt: on

        # horizontal margin
        self._inter_column_margin: Decimal = Decimal(self._page_width) * Decimal(0.05)
        if inter_column_margin is not None:
            self._inter_column_margin = inter_column_margin
        if horizontal_margin is None:
            self._horizontal_margin = self._page_width * Decimal(0.1)
        else:
            assert horizontal_margin >= 0
            self._horizontal_margin = horizontal_margin

        # vertical margin
        if vertical_margin is None:
            self._vertical_margin_top = self._page_height * Decimal(0.1)
            self._vertical_margin_bottom = self._vertical_margin_top
        else:
            assert vertical_margin >= 0
            self._vertical_margin_top = vertical_margin
            self._vertical_margin_bottom = vertical_margin

        self._number_of_columns = Decimal(number_of_columns)
        self._column_width = (
            self._page_width
            - Decimal(2) * self._horizontal_margin
            - Decimal(number_of_columns - 1) * self._inter_column_margin
        ) / Decimal(number_of_columns)

        # previous element information
        self._previous_element: typing.Optional[LayoutElement] = None
        self._current_column_index = Decimal(0)

    def _get_margin_between_elements(
        self, previous_element: LayoutElement, element: LayoutElement
    ) -> Decimal:
        if previous_element is None:
            return Decimal(0)

        # text elements
        if isinstance(previous_element, ChunkOfText) and isinstance(
            element, ChunkOfText
        ):
            return max(
                previous_element.get_font_size() * Decimal(1.2),
                element.get_font_size() * Decimal(1.2),
            )
        if isinstance(previous_element, ChunkOfText):
            return previous_element.get_font_size() * Decimal(1.2)
        if isinstance(element, ChunkOfText):
            return element.get_font_size() * Decimal(1.2)

        # default
        return Decimal(5)

    def switch_to_next_column(self) -> "PageLayout":
        """
        This function forces this PageLayout to move to the next column on the Page
        """
        self._current_column_index += Decimal(1)
        if self._current_column_index == self._number_of_columns:
            return self.switch_to_next_page()
        assert self._page_height
        self._previous_element = None
        return self

    def switch_to_next_page(self) -> "PageLayout":
        """
        This function forces this PageLayout to move to the next Page
        """
        self._current_column_index = Decimal(0)
        assert self._page_height
        self._previous_element = None

        # find Document
        doc = self.get_page().get_root()  # type: ignore[attr-defined]
        assert isinstance(doc, Document)

        # create new Page
        assert self._page_width
        assert self._page_height
        new_page = Page(width=self._page_width, height=self._page_height)
        self._page = new_page
        doc.add_page(new_page)

        # return
        return self

    def add(self, layout_element: LayoutElement) -> "PageLayout":
        """
        This method adds a `LayoutElement` to the current `Page`.
        """
        if self._current_column_index >= self._number_of_columns:
            return self

        # previous element is used to determine the paragraph spacing
        assert self._page_height is not None
        assert self._page_width is not None
        previous_element_margin_bottom: Decimal = Decimal(0)
        previous_element_y = self._page_height - self._vertical_margin_top
        if self._previous_element is not None:
            previous_element_y = (
                self._previous_element.get_previous_layout_box().get_y()
            )
            previous_element_margin_bottom = self._previous_element.get_margin_bottom()

        # calculate next available rectangle
        available_height: Decimal = (
            previous_element_y
            - self._vertical_margin_bottom
            - self._get_margin_between_elements(self._previous_element, layout_element)
            - max(previous_element_margin_bottom, layout_element.get_margin_top())
            - layout_element.get_margin_bottom()
        )

        # switch to new column if needed
        assert self._page_height
        if available_height < 0:
            self.switch_to_next_column()
            return self.add(layout_element)

        # fmt: off
        available_space: Rectangle = Rectangle(
            self._horizontal_margin + self._current_column_index * (self._column_width + self._inter_column_margin) + layout_element.get_margin_left(),
            self._vertical_margin_bottom + layout_element.get_margin_bottom(),
            self._column_width - layout_element.get_margin_right() - layout_element.get_margin_left(),
            available_height
        )
        # fmt: on

        # get layout_box
        layout_rect = layout_element.get_layout_box(available_space)
        if layout_rect.get_height() > available_space.get_height():
            if self._previous_element is not None:
                self.switch_to_next_column()
                return self.add(layout_element)
            else:
                assert False, (
                    "%s is too tall to fit inside column / page."
                    % layout_element.__class__.__name__
                )

        # content that is too large for the page should not be attempted to be laid out again
        if round(layout_rect.get_width(), 2) > round(self._column_width, 2):
            assert False, (
                "%s is too wide to fit inside column / page."
                % layout_element.__class__.__name__
            )

        # switch to next column
        if layout_rect.y < self._vertical_margin_bottom:
            self.switch_to_next_column()
            return self.add(layout_element)

        # paint
        layout_element.paint(self._page, available_space)

        # store previous
        self._previous_element = layout_element
        self._previous_element_layout_rect = layout_rect

        # return
        return self


class SingleColumnLayout(MultiColumnLayout):
    """
    This implementation of PageLayout adds left/right/top/bottom margins to a Page
    and lays out the content on the Page as if there were was a single column to flow text, images, etc into.
    Once this column is full, the next page is automatically created.
    """

    def __init__(
        self,
        page: Page,
        horizontal_margin: typing.Optional[Decimal] = None,
        vertical_margin: typing.Optional[Decimal] = None,
    ):
        super(SingleColumnLayout, self).__init__(
            page,
            number_of_columns=1,
            horizontal_margin=horizontal_margin,
            vertical_margin=vertical_margin,
        )
        self._inter_column_margin = Decimal(0)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
    Beginning with PDF 1.5, cross-reference information may be stored in a cross-reference stream instead of in a
    cross-reference table.
"""
import io
from decimal import Decimal
from typing import Optional, Union

from borb.io.filter.stream_decode_util import decode_stream
from borb.io.read.tokenize.high_level_tokenizer import HighLevelTokenizer
from borb.io.read.types import Dictionary, List, Name, Reference, Stream
from borb.pdf.xref.xref import XREF


class StreamXREF(XREF):
    """
    Beginning with PDF 1.5, cross-reference information may be stored in a cross-reference stream instead of in a
    cross-reference table. Cross-reference streams provide the following advantages:
    • A more compact representation of cross-reference information
    • The ability to access compressed objects that are stored in object streams (see 7.5.7, "Object Streams")
    and to allow new cross-reference entry types to be added in the future
    Cross-reference streams are stream objects (see 7.3.8, "Stream Objects"), and contain a dictionary and a data
    stream. Each cross-reference stream contains the information equivalent to the cross-reference table
     (see 7.5.4, "Cross-Reference Table") and trailer (see 7.5.5, "File Trailer") for one cross-reference section.
    """

    def __init__(self, initial_offset: Optional[int] = None):
        super().__init__()
        self._initial_offset = initial_offset

    def read(
        self,
        io_source: Union[io.BufferedIOBase, io.RawIOBase, io.BytesIO],
        tokenizer: HighLevelTokenizer,
        initial_offset: Optional[int] = None,
    ) -> "XREF":
        """
        This method attempts to read a stream XREF from the given io_source.
        It will either throw an exception, or return this XREF
        """

        if initial_offset is not None:
            io_source.seek(initial_offset)
        else:
            self._seek_to_xref_token(io_source, tokenizer)

        xref_stream = tokenizer.read_object()
        assert isinstance(xref_stream, Stream)

        # check widths
        assert "W" in xref_stream
        assert all(
            [
                isinstance(xref_stream["W"][x], Decimal)
                for x in range(0, len(xref_stream["W"]))
            ]
        )
        # decode widths
        widths = [int(xref_stream["W"][x]) for x in range(0, len(xref_stream["W"]))]
        total_entry_width = sum(widths)

        # parent
        document = self.get_root()  # type: ignore [attr-defined]

        # list of references
        indirect_references = [
            Reference(
                object_number=0,
                generation_number=65535,
                is_in_use=False,
                document=document,
            )
        ]

        # check size
        assert "Size" in xref_stream
        assert isinstance(xref_stream["Size"], Decimal)

        # get size
        number_of_objects = int(xref_stream["Size"])

        # index
        index = []
        if "Index" in xref_stream:
            index = xref_stream["Index"]
            assert isinstance(index, List)
            assert len(index) % 2 == 0
            assert isinstance(index[0], Decimal)
            assert isinstance(index[1], Decimal)
        else:
            index = [Decimal(0), Decimal(number_of_objects)]

        # apply filters
        xref_stream = decode_stream(xref_stream)

        # read every range specified in /Index
        xref_stream_decoded_bytes = xref_stream["DecodedBytes"]
        for idx in range(0, len(index), 2):
            start = int(index[idx])
            length = int(index[idx + 1])

            bptr = 0
            for i in range(0, length):

                # object number
                object_number = start + i

                # read type
                type = 1
                if widths[0] > 0:
                    type = 0
                    for j in range(0, widths[0]):
                        type = (type << 8) + (xref_stream_decoded_bytes[bptr] & 0xFF)
                        bptr += 1

                # read field 2
                field2 = 0
                for j in range(0, widths[1]):
                    field2 = (field2 << 8) + (xref_stream_decoded_bytes[bptr] & 0xFF)
                    bptr += 1

                # read field 3
                field3 = 0
                for j in range(0, widths[2]):
                    field3 = (field3 << 8) + (xref_stream_decoded_bytes[bptr] & 0xFF)
                    bptr += 1

                # check type
                assert type in [0, 1, 2]

                pdf_indirect_reference = None
                if type == 0:
                    # type      :The type of this entry, which shall be 0. Type 0 entries define
                    # the linked list of free objects (corresponding to f entries in a
                    # cross-reference table).
                    # field2    : The object number of the next free object
                    # field3    : The generation number to use if this object number is used again
                    pdf_indirect_reference = Reference(
                        document=document,
                        object_number=object_number,
                        byte_offset=field2,
                        generation_number=field3,
                        is_in_use=False,
                    )

                if type == 1:
                    # Type      : The type of this entry, which shall be 1. Type 1 entries define
                    # objects that are in use but are not compressed (corresponding
                    # to n entries in a cross-reference table).
                    # field2    : The byte offset of the object, starting from the beginning of the
                    # file.
                    # field3    : The generation number of the object. Default value: 0.
                    pdf_indirect_reference = Reference(
                        document=document,
                        object_number=object_number,
                        byte_offset=field2,
                        generation_number=field3,
                    )

                if type == 2:
                    # Type      : The type of this entry, which shall be 2. Type 2 entries define
                    # compressed objects.
                    # field2    : The object number of the object stream in which this object is
                    # stored. (The generation number of the object stream shall be
                    # implicitly 0.)
                    # field3    : The index of this object within the object stream.
                    pdf_indirect_reference = Reference(
                        document=document,
                        object_number=object_number,
                        generation_number=0,
                        parent_stream_object_number=field2,
                        index_in_parent_stream=field3,
                    )

                assert pdf_indirect_reference is not None

                # append
                existing_indirect_ref = next(
                    iter(
                        [
                            x
                            for x in indirect_references
                            if x.object_number is not None
                            and x.object_number == Decimal(object_number)
                        ]
                    ),
                    None,
                )
                ref_is_in_reading_state = (
                    existing_indirect_ref is not None
                    and existing_indirect_ref.is_in_use
                    and existing_indirect_ref.generation_number
                    == pdf_indirect_reference.generation_number
                )
                ref_is_first_encountered = existing_indirect_ref is None or (
                    not ref_is_in_reading_state
                    and existing_indirect_ref.document is None
                )

                if ref_is_first_encountered:
                    assert pdf_indirect_reference is not None
                    indirect_references.append(pdf_indirect_reference)
                elif ref_is_in_reading_state:
                    assert existing_indirect_ref is not None
                    assert pdf_indirect_reference is not None
                    existing_indirect_ref.index_in_parent_stream = (
                        pdf_indirect_reference.index_in_parent_stream
                    )
                    existing_indirect_ref.parent_stream_object_number = (
                        pdf_indirect_reference.parent_stream_object_number
                    )

        # add section
        for r in indirect_references:
            self.add(r)

        # initialize trailer
        self[Name("Trailer")] = Dictionary()
        for k, v in xref_stream.items():
            self[Name("Trailer")][k] = v
        self[Name("Trailer")].set_parent(self)

        # return
        return self



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
    This module contains the build information for this project
"""
import sys

import setuptools

# open readme
with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

# requirements
required = [
    "fonttools>=4.22.1",                        # TTF fonts
    "Pillow>=7.1.0",                            # image processing
    "python-barcode>=0.13.1",                   # generating barcodes
    "qrcode[pil]>=6.1",                         # generating QR codes
    "requests>=2.24.0",                         # generating images from a URL
    "setuptools>=51.1.1",                       # standard
]

dependency_links = [
    "http://github.com/ojii/pymaging/tarball/master",
    "http://github.com/ojii/pymaging-png/tarball/master",
]

setuptools.setup(
    name="borb",
    version="2.1.3",
    author="Joris Schellekens",
    author_email="joris.schellekens.1989@gmail.com",
    description="borb is a library for reading, creating and manipulating PDF files in python.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/jorisschellekens/borb",
    packages=setuptools.find_packages(include=["borb", "borb.*"]),
    include_package_data=True,
    install_requires=required,
    dependency_links=dependency_links,
    python_requires=">=3.6",
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Begin a marked-content sequence with an associated property list,
terminated by a balancing EMC operator. tag shall be a name object
indicating the role or significance of the sequence. properties shall be
either an inline dictionary containing the property list or a name object
associated with it in the Properties subdictionary of the current resource
dictionary (see 14.6.2, “Property Lists”).
"""

import typing

from borb.io.read.types import AnyPDFType, Name
from borb.pdf.canvas.operator.canvas_operator import CanvasOperator


class BeginMarkedContentWithPropertyList(CanvasOperator):
    """
    Begin a marked-content sequence with an associated property list,
    terminated by a balancing EMC operator. tag shall be a name object
    indicating the role or significance of the sequence. properties shall be
    either an inline dictionary containing the property list or a name object
    associated with it in the Properties subdictionary of the current resource
    dictionary (see 14.6.2, “Property Lists”).
    """

    def __init__(self):
        super().__init__("BDC", 2)

    def invoke(
        self,
        canvas_stream_processor: "CanvasStreamProcessor",  # type: ignore [name-defined]
        operands: typing.List[AnyPDFType] = [],
        event_listeners: typing.List["EventListener"] = [],  # type: ignore [name-defined]
    ) -> None:
        """
        Invoke the BDC operator
        """
        assert isinstance(operands[0], Name), "Operand 0 of BDC must be a Name"
        canvas = canvas_stream_processor.get_canvas()
        canvas.marked_content_stack.append(operands[0])



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
A metadata stream may be attached to a document through the Metadata entry in the document catalogue
(see 7.7.2, “Document Catalog”). The metadata framework provides a date stamp for metadata expressed in
the framework. If this date stamp is equal to or later than the document modification date recorded in the
document information dictionary, the metadata stream shall be taken as authoritative. If, however, the
document modification date recorded in the document information dictionary is later than the metadata
stream’s date stamp, the document has likely been saved by a writer that is not aware of metadata streams. In
this case, information stored in the document information dictionary shall be taken to override any semantically
equivalent items in the metadata stream. In addition, PDF document components represented as a stream or
dictionary may have a Metadata entry (see Table 316).
"""
import io
import logging
import typing
import xml.etree.ElementTree as ET
from typing import Any, Optional, Union

from borb.io.read.object.stream_transformer import StreamTransformer
from borb.io.read.transformer import ReadTransformerState
from borb.io.read.types import AnyPDFType, Element, Stream
from borb.pdf.canvas.event.event_listener import EventListener

logger = logging.getLogger(__name__)


class XMPMetadataTransformer(StreamTransformer):
    """
    A metadata stream may be attached to a document through the Metadata entry in the document catalogue
    (see 7.7.2, “Document Catalog”). The metadata framework provides a date stamp for metadata expressed in
    the framework. If this date stamp is equal to or later than the document modification date recorded in the
    document information dictionary, the metadata stream shall be taken as authoritative. If, however, the
    document modification date recorded in the document information dictionary is later than the metadata
    stream’s date stamp, the document has likely been saved by a writer that is not aware of metadata streams. In
    this case, information stored in the document information dictionary shall be taken to override any semantically
    equivalent items in the metadata stream. In addition, PDF document components represented as a stream or
    dictionary may have a Metadata entry (see Table 316).
    """

    def can_be_transformed(
        self, object: Union[io.BufferedIOBase, io.RawIOBase, io.BytesIO, AnyPDFType]
    ) -> bool:
        """
        This function returns True if the object to be converted represents an XML element
        """
        return (
            isinstance(object, Stream)
            and "Type" in object
            and object["Type"] == "Metadata"
            and "Subtype" in object
            and object["Subtype"] == "XML"
        )

    def transform(
        self,
        object_to_transform: Union[io.BufferedIOBase, io.RawIOBase, AnyPDFType],
        parent_object: Any,
        context: Optional[ReadTransformerState] = None,
        event_listeners: typing.List[EventListener] = [],
    ) -> Any:
        """
        This function reads an XML Element from a byte stream
        """

        # delegate to super (ReadStreamTransformer)
        out_value = super(XMPMetadataTransformer, self).transform(
            object_to_transform=object_to_transform,
            parent_object=parent_object,
            context=context,
            event_listeners=event_listeners,
        )

        # parse XML
        assert out_value is not None
        assert isinstance(out_value, Stream)
        assert "DecodedBytes" in out_value
        xml_root_out = None
        try:
            xml_root_orig = ET.fromstring(out_value["DecodedBytes"].decode("latin1"))

            # make copy so that we can add attributes like parent and listeners
            xml_root_out = Element(xml_root_orig.tag)
            xml_root_out.set_parent(parent_object)  # type: ignore [attr-defined]
            for e in xml_root_orig:
                xml_root_out.append(e)

        except Exception as ex:
            logger.warning("Unable to process XMP meta-data")

        # return
        return xml_root_out



# !/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This module contains everything needed to perform low-level tokenization against PDF syntax.
Low-level tokenization aims to separate numbers, strings, names, comments, start of dictionary, start of array, etc
The high-level tokenizer will use this first pass to then build complex objects (streams, dictionaries, etc)
"""
import re
from typing import Optional

from borb.io.read.tokenize.low_level_tokenizer import LowLevelTokenizer, TokenType
from borb.io.read.types import (
    AnyPDFType,
    Boolean,
    CanvasOperatorName,
    Decimal,
    Dictionary,
    HexadecimalString,
    List,
    Name,
    Reference,
    Stream,
    String,
)


class HighLevelTokenizer(LowLevelTokenizer):
    """
    In computer science, lexical analysis, lexing or tokenization is the process of converting a sequence of characters
    (such as in a computer program or web page) into a sequence of tokens (strings with an assigned and thus identified meaning).
    A program that performs lexical analysis may be termed a lexer, tokenizer, or scanner,
    although scanner is also a term for the first stage of a lexer.
    A lexer is generally combined with a parser, which together analyze the syntax of programming languages, web pages,
    and so forth.
    """

    def read_array(self) -> List:
        """
        This function processes the next tokens and returns a List.
        It fails and throws various errors if the next tokens do not represent a List.
        """
        token = self.next_non_comment_token()
        assert token is not None
        assert token.get_token_type() == TokenType.START_ARRAY
        out = List()

        while True:
            token = self.next_non_comment_token()
            assert token is not None
            if token.get_token_type() == TokenType.END_ARRAY:
                break
            assert token.get_token_type() != TokenType.END_DICT

            # go back
            self.seek(token.get_byte_offset())

            # read
            obj = self.read_object()

            # append
            out.append(obj)

        # return
        return out

    def read_dictionary(self) -> Dictionary:
        """
        This function processes the next tokens and returns a Dictionary.
        It fails and throws various errors if the next tokens do not represent a Dictionary.
        """
        token = self.next_non_comment_token()
        assert token is not None
        assert token.get_token_type() == TokenType.START_DICT

        out_dict = Dictionary()
        while True:

            # attempt to read name token
            token = self.next_non_comment_token()
            assert token is not None
            if token.get_token_type() == TokenType.END_DICT:
                break
            assert token.get_token_type() == TokenType.NAME

            # store name
            name = Name(token.get_text()[1:])

            # attempt to read value
            value = self.read_object()
            assert value is not None

            # store in dict object
            if name is not None:
                out_dict[name] = value

        return out_dict

    def read_indirect_object(self) -> Optional[AnyPDFType]:
        """
        This function processes the next tokens and returns an AnyPDFType.
        It fails and throws various errors if the next tokens do not represent an indirect pdf object.
        """

        # read object number
        token = self.next_non_comment_token()
        assert token is not None
        byte_offset = token.get_byte_offset()
        if token.get_token_type() != TokenType.NUMBER or not re.match(
            "^[0-9]+$", token.get_text()
        ):
            self.seek(byte_offset)
            return None
        object_number = int(token.get_text())

        # read generation number
        token = self.next_non_comment_token()
        assert token is not None
        if token.get_token_type() != TokenType.NUMBER or not re.match(
            "^[0-9]+$", token.get_text()
        ):
            self.seek(byte_offset)
            return None
        generation_number = int(token.get_text())

        # read 'obj'
        token = self.next_non_comment_token()
        assert token is not None
        if token.get_token_type() != TokenType.OTHER or token.get_text() != "obj":
            self.seek(byte_offset)
            return None

        # read obj
        value = self.read_object()
        if value is not None:
            value.set_reference(  # type: ignore[union-attr]
                Reference(
                    object_number=object_number, generation_number=generation_number
                )
            )

        # return
        return value

    def read_indirect_reference(self) -> Optional[Reference]:
        """
        This function processes the next tokens and returns an indirect reference.
        It fails and throws various errors if the next tokens do not represent an indirect reference.
        """

        # read object number
        token = self.next_non_comment_token()
        assert token is not None
        byte_offset = token.get_byte_offset()
        if token.get_token_type() != TokenType.NUMBER or not re.match(
            "^[0-9]+$", token.get_text()
        ):
            self.seek(byte_offset)
            return None
        object_number = int(token.get_text())

        # read generation number
        token = self.next_non_comment_token()
        assert token is not None
        if token.get_token_type() != TokenType.NUMBER or not re.match(
            "^[0-9]+$", token.get_text()
        ):
            self.seek(byte_offset)
            return None
        generation_number = int(token.get_text())

        # read 'R'
        token = self.next_non_comment_token()
        assert token is not None
        if token.get_token_type() != TokenType.OTHER or token.get_text() != "R":
            self.seek(byte_offset)
            return None

        # return
        return Reference(
            object_number=object_number,
            generation_number=generation_number,
        )

    def read_object(self, xref: Optional["XREF"] = None) -> Optional[AnyPDFType]:  # type: ignore [name-defined]
        """
        This function processes the next tokens and returns an AnyPDFType.
        It fails and throws various errors if the next tokens do not represent a pdf object.
        """
        token = self.next_non_comment_token()
        if token is None or len(token.get_text()) == 0:
            return None

        if token.get_token_type() == TokenType.START_DICT:
            self.seek(token.get_byte_offset())  # go to start of dictionary
            return self.read_dictionary()

        if token.get_token_type() == TokenType.START_ARRAY:
            self.seek(token.get_byte_offset())  # go to start of array
            return self.read_array()

        # <number> <number> "R"
        if token.get_token_type() == TokenType.NUMBER:
            self.seek(token.get_byte_offset())  # go to start of indirect reference
            potential_indirect_reference = self.read_indirect_reference()
            if potential_indirect_reference is not None:
                return potential_indirect_reference

        # <number> <number> "obj"
        # <<dictionary>>
        # "stream"
        # <bytes>
        # "endstream"
        if token.get_token_type() == TokenType.NUMBER:
            self.seek(token.get_byte_offset())
            potential_stream = self.read_stream(xref)
            if potential_stream is not None:
                return potential_stream

        # <number> <number> "obj"
        if token.get_token_type() == TokenType.NUMBER:
            self.seek(token.get_byte_offset())
            potential_indirect_object = self.read_indirect_object()
            if potential_indirect_object is not None:
                return potential_indirect_object

        # numbers
        if token.get_token_type() == TokenType.NUMBER:
            self.seek(self.tell() + len(token.get_text()))
            return Decimal(Decimal(token.get_text()))

        # boolean
        if token.get_token_type() == TokenType.OTHER and token.get_text() in [
            "true",
            "false",
        ]:
            return Boolean(token.get_text() == "true")

        # canvas operators
        if (
            token.get_token_type() == TokenType.OTHER
            and token.get_text() in CanvasOperatorName.VALID_NAMES
        ):
            return CanvasOperatorName(token.get_text())

        # names
        if token.get_token_type() == TokenType.NAME:
            return Name(token.get_text()[1:])

        # literal strings and hex strings
        if token.get_token_type() in [TokenType.STRING, TokenType.HEX_STRING]:
            if token.get_token_type() == TokenType.STRING:
                # TODO: we should pass the bytes rather than the text
                return String(token.get_text()[1:-1])
            else:
                return HexadecimalString(token.get_text()[1:-1])

        # default
        return None

    def read_stream(self, xref: Optional["XREF"] = None) -> Optional[Stream]:  # type: ignore [name-defined]
        """
        This function processes the next tokens and returns a Stream.
        It fails and throws various errors if the next tokens do not represent a Stream.
        """
        byte_offset = self.tell()

        # attempt to read <number> <number> obj
        # followed by dictionary
        stream_dictionary = self.read_indirect_object()
        if stream_dictionary is None or not isinstance(stream_dictionary, dict):
            self.seek(byte_offset)
            return None

        # attempt to read keyword "stream"
        stream_token = self.next_non_comment_token()
        assert stream_token is not None
        if (
            stream_token.get_token_type() != TokenType.OTHER
            or stream_token.get_text() != "stream"
        ):
            self.seek(byte_offset)
            return None

        # process /Length
        assert "Length" in stream_dictionary
        length_of_stream = stream_dictionary["Length"]
        if isinstance(length_of_stream, Reference):
            if xref is None:
                raise RuntimeError(
                    "unable to process reference /Length when no XREF is given"
                )
            pos_before = self.tell()
            length_of_stream = int(
                xref.get_object(length_of_stream, src=self._io_source, tok=self)
            )
            self.seek(pos_before)

        # process newline
        ch = self._next_byte()
        assert ch in [b"\r", b"\n"]
        if ch == b"\r":
            ch = self._next_byte()
            assert ch == b"\n"

        bytes = self._io_source.read(int(length_of_stream))

        # attempt to read token "endstream"
        end_of_stream_token = self.next_non_comment_token()
        assert end_of_stream_token is not None
        assert end_of_stream_token.get_token_type() == TokenType.OTHER
        assert end_of_stream_token.get_text() == "endstream"

        # set Bytes
        stream_dictionary[Name("Bytes")] = bytes

        # return
        output: Stream = Stream()
        for k, v in stream_dictionary.items():
            output[k] = v
        return output



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Some categories of objects in a PDF file can be referred to by name rather than by object reference. The
correspondence between names and objects is established by the document’s name dictionary (PDF 1.2),
located by means of the Names entry in the document’s catalog (see 7.7.2, "Document Catalog"). Each entry in
this dictionary designates the root of a name tree (see 7.9.6, "Name Trees") defining names for a particular
category of objects.

A name tree serves a similar purpose to a dictionary—associating keys and values—but by different means.
A name tree differs from a dictionary in the following important ways:
- Unlike the keys in a dictionary, which are name objects, those in a name tree are strings.
- The keys are ordered.
- The values associated with the keys may be objects of any type. Stream objects shall be specified by
indirect object references (7.3.8, "Stream Objects"). The dictionary, array, and string objects should be
specified by indirect object references, and other PDF objects (nulls, numbers, booleans, and names)
should be specified as direct objects.
- The data structure can represent an arbitrarily large collection of key-value pairs, which can be looked up
efficiently without requiring the entire data structure to be read from the PDF file. (In contrast, a dictionary
can be subject to an implementation limit on the number of entries it can contain.)
"""
import typing

from borb.io.read.types import Dictionary, Name, List, String


class NameTree:
    """
    A name tree is similar to a dictionary that associates keys and values but the keys in a name tree are strings and are ordered
    """

    def __init__(self, document: Dictionary, name: Name):
        self._document: Dictionary = document
        self._name: Name = name

    def put(self, key: str, value: typing.Any) -> "NameTree":
        """
        This function adds a key/value pair in this NameTree
        :param key:     the key
        :param value:   the value
        :return:        self
        """
        assert "XRef" in self._document, "No XREF found in this PDF"
        assert (
            "Trailer" in self._document["XRef"]
        ), "No /Trailer dictionary found in the XREF"
        assert (
            "Root" in self._document["XRef"]["Trailer"]
        ), "No /Root dictionary found in the /Trailer"
        root = self._document["XRef"]["Trailer"]["Root"]

        # set up /Names dictionary
        if "Names" not in root:
            root[Name("Names")] = Dictionary()
        names = root["Names"]

        # set up /JavaScript
        if self._name not in names:
            names[self._name] = Dictionary()
            names[self._name][Name("Kids")] = List()

        # find parent
        parent = names[self._name]
        while "Kids" in parent:
            for k in parent["Kids"]:
                lower_limit = str(k["Limits"][0])
                upper_limit = str(k["Limits"][1])
                if lower_limit == upper_limit:
                    continue
                if lower_limit < key < upper_limit:
                    parent = k
                    break
            break

        # insert as new child, or replace existing child?
        if (
            len([x for x in parent["Kids"] if x["Limits"][0] == x["Limits"][1] == key])
            == 0
        ):
            self._put_new(parent, key, value)
        else:
            self._put_existing(parent, key, value)

        # return
        return self

    def _put_new(self, parent: Dictionary, key: str, value: typing.Any):

        kid = Dictionary()
        kid[Name("F")] = String(key)
        kid[Name("Limits")] = List()
        for _ in range(0, 2):
            kid["Limits"].append(String(key))
        kid[Name("Names")] = List()
        kid[Name("Names")].append(String(key))

        # build leaf /Filespec dictionary
        if self._name == "EmbeddedFiles":
            kid[Name("Names")].append(value)
            kid[Name("Type")] = Name("EF")

        # build leaf /JavaScript dictionary
        if self._name == "JavaScript":
            kid[Name("Names")].append(value)

        # append
        parent["Kids"].append(kid)

    def _put_existing(self, parent: Dictionary, key: str, value: typing.Any):
        # TODO
        pass

    #
    # dictionary methods
    #

    def keys(self) -> typing.List[String]:
        """
        This function returns the keys in this NameTree
        :return:    the keys in this NameTree
        """
        return [k for k, v in self.items()]

    def values(self) -> typing.List[typing.Any]:
        """
        This function returns the values in this NameTree
        :return:    the values in this NameTree
        """
        return [v for k, v in self.items()]

    def items(self) -> typing.Iterable[typing.Tuple[String, typing.Any]]:
        """
        This function returns all key/value pairs in this NameTree
        :return:    all key/value pairs in this NameTree
        """
        assert "XRef" in self._document, "No XREF found in this PDF"
        assert (
            "Trailer" in self._document["XRef"]
        ), "No /Trailer dictionary found in the XREF"
        assert (
            "Root" in self._document["XRef"]["Trailer"]
        ), "No /Root dictionary found in the /Trailer"
        root = self._document["XRef"]["Trailer"]["Root"]

        # set up /Names dictionary
        if "Names" not in root:
            root[Name("Names")] = Dictionary()
        names = root["Names"]

        nodes_to_visit = [names[self._name]]
        keys = []
        values = []
        while len(nodes_to_visit) > 0:
            n = nodes_to_visit[0]
            nodes_to_visit.pop(0)
            if "Kids" in n:
                for k in n["Kids"]:
                    nodes_to_visit.append(k)
            if "Limits" in n:
                lower_limit = str(n["Limits"][0])
                upper_limit = str(n["Limits"][1])
                if upper_limit == lower_limit:
                    keys.append(upper_limit)
                    values.append(n["Names"][1])

        # return
        return zip(keys, values)

    #
    # utility functions
    #

    def _get_root_or_empty(self):
        assert "XRef" in self._document, "No XREF found in this PDF"
        assert (
            "Trailer" in self._document["XRef"]
        ), "No /Trailer dictionary found in the XREF"
        assert (
            "Root" in self._document["XRef"]["Trailer"]
        ), "No /Root dictionary found in the /Trailer"
        root = self._document["XRef"]["Trailer"]["Root"]
        return root.get(Name("Names"), Dictionary())

    def __len__(self):
        return len(self._get_root_or_empty())



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This implementation of  EventListener scans the pages of a PDF for tables.
It recognizes a table whenever a substantial amount of intersecting (or connected) horizontal/vertical lines occur.
The recognized grid must have at least 2 cells otherwise it could just be a piece of text in a box.
"""
import logging
import math
import typing
from decimal import Decimal

from borb.datastructure.disjoint_set import disjointset
from borb.pdf.canvas.event.begin_page_event import BeginPageEvent
from borb.pdf.canvas.event.chunk_of_text_render_event import ChunkOfTextRenderEvent
from borb.pdf.canvas.event.end_page_event import EndPageEvent
from borb.pdf.canvas.event.event_listener import Event, EventListener
from borb.pdf.canvas.event.line_render_event import LineRenderEvent
from borb.pdf.canvas.geometry.line_segment import LineSegment
from borb.pdf.canvas.geometry.rectangle import Rectangle
from borb.pdf.canvas.layout.table.flexible_column_width_table import (
    FlexibleColumnWidthTable,
)
from borb.pdf.canvas.layout.table.table import Table, TableCell
from borb.pdf.canvas.layout.text.paragraph import Paragraph

logger = logging.getLogger(__name__)


class TableDetectionByLines(EventListener):
    """
    This implementation of  EventListener scans the pages of a PDF for tables.
    It recognizes a table whenever a substantial amount of intersecting (or connected) horizontal/vertical lines occur.
    The recognized grid must have at least 2 cells otherwise it could just be a piece of text in a box.
    """

    def __init__(self):
        self._current_page_number: int = -1
        self._lines_per_page: typing.Dict[int, typing.List[LineSegment]] = {}
        self._tables_per_page: typing.Dict[int, typing.List[Table]] = {}
        self._text_render_events_per_page: typing.Dict[
            int, typing.List[ChunkOfTextRenderEvent]
        ] = {}

    @staticmethod
    def _dist(x0: Decimal, y0: Decimal, x1: Decimal, y1: Decimal) -> Decimal:
        return Decimal(math.sqrt((x0 - x1) ** 2 + (y0 - y1) ** 2))

    def get_table_bounding_boxes_for_page(
        self, page_number: int
    ) -> typing.List[Rectangle]:
        """
        This function returns the bounding boxes (as Rectangle objects) of each Table
        that was recognized on the given page.
        """
        ZERO: Decimal = Decimal(0)
        return [
            x.get_previous_layout_box() or Rectangle(ZERO, ZERO, ZERO, ZERO)
            for x in self._tables_per_page.get(page_number, [])
            if x.get_previous_layout_box() is not None
        ]

    def get_tables_for_page(self, page_number: int) -> typing.List[Table]:
        """
        This function returns each Table that was recognized on the given page.
        """
        return self._tables_per_page.get(page_number, [])

    def _event_occurred(self, event: Event) -> None:

        # BeginPageEvent
        if isinstance(event, BeginPageEvent):
            self._current_page_number += 1
            self._lines_per_page[self._current_page_number] = []
            self._tables_per_page[self._current_page_number] = []
            self._text_render_events_per_page[self._current_page_number] = []

        # ChunkOfTextRenderEvent
        if isinstance(event, ChunkOfTextRenderEvent):
            self._text_render_events_per_page[self._current_page_number].append(event)

        # LineRenderEvent
        if isinstance(event, LineRenderEvent):
            ls: LineSegment = event.get_line_segment()
            if abs(ls.x0 - ls.x1) > Decimal(1) and abs(ls.y0 - ls.y1) > Decimal(1):
                return
            if TableDetectionByLines._dist(ls.x0, ls.y0, ls.x1, ls.y1) < Decimal(1):
                return
            self._lines_per_page[self._current_page_number].append(
                event.get_line_segment()
            )

        # EndPageEvent
        if isinstance(event, EndPageEvent):
            ds = disjointset()
            for l in self._lines_per_page[self._current_page_number]:
                ds.add(l)
            for l0 in self._lines_per_page[self._current_page_number]:
                for l1 in self._lines_per_page[self._current_page_number]:

                    if l0 == l1:
                        continue
                    if ds.find(l0) == ds.find(l1):
                        continue

                    # two LineSegment objects share (at least) 1 point
                    # fmt: off
                    if TableDetectionByLines._dist(l0.x0, l0.y0, l1.x0, l1.y0) < Decimal(1):
                        ds.union(l0, l1)
                        continue
                    if TableDetectionByLines._dist(l0.x0, l0.y0, l1.x1, l1.y1) < Decimal(1):
                        ds.union(l0, l1)
                        continue
                    if TableDetectionByLines._dist(l0.x1, l0.y1, l1.x0, l1.y0) < Decimal(1):
                        ds.union(l0, l1)
                        continue
                    if TableDetectionByLines._dist(l0.x1, l0.y1, l1.x1, l1.y1) < Decimal(1):
                        ds.union(l0, l1)
                        continue
                    # fmt: on

                    # two LineSegment objects (A, B) such that 1 point of A lies on B
                    # fmt: off
                    if abs(l0.x0 - l0.x1) <= Decimal(1)                                             \
                        and (abs(l0.x0 - l1.x0) <= Decimal(1) or abs(l0.x0 - l1.x1) <= Decimal(1))  \
                        and abs(l1.y0 - l1.y1) <= Decimal(1)                                        \
                        and min(l0.y0, l0.y1) <= l1.y0 <= max(l0.y0, l0.y1):
                        ds.union(l0, l1)
                    if abs(l0.y0 - l0.y1) <= Decimal(1)                                             \
                        and (abs(l0.y0 - l1.y0) <= Decimal(1) or abs(l0.y0 - l1.y1) <= Decimal(1))  \
                        and abs(l1.x0 - l1.x1) <= Decimal(1)                                        \
                        and min(l0.x0, l0.x1) <= l1.x0 <= max(l0.x0, l0.x1):
                        ds.union(l0, l1)
                    # fmt: on

                    # two LineSegment objects intersect at right angle
                    # TODO

            # extract clusters
            clusters_of_lines: typing.Dict[LineSegment, typing.List[LineSegment]] = {}
            for l in ds:
                if ds.find(l) not in clusters_of_lines:
                    clusters_of_lines[ds.find(l)] = []
                clusters_of_lines[ds.find(l)].append(l)

            # process all lines per table individually
            for _, v in clusters_of_lines.items():
                r, c = self._determine_number_of_rows_and_columns(v)
                if r * c >= 2:

                    # determine table
                    table: Table = self._determine_table_cell_boundaries(v)
                    table._previous_layout_box = self._determine_table_bounding_box(v)
                    table._previous_paint_box = table._previous_layout_box

                    # store
                    self._tables_per_page[self._current_page_number].append(table)

    def _determine_number_of_rows_and_columns(
        self, lines_in_table: typing.List[LineSegment]
    ) -> typing.Tuple[int, int]:

        # keep track of unique xs / ys (to derive number of rows/cols)
        unique_xs: typing.Set[int] = set()
        unique_ys: typing.Set[int] = set()

        for l in lines_in_table:
            unique_xs.add(int(l.x0))
            unique_xs.add(int(l.x1))
            unique_ys.add(int(l.y0))
            unique_ys.add(int(l.y1))

        # determine number of rows/cols
        number_of_rows: int = len(unique_ys) - 1
        number_of_cols: int = len(unique_xs) - 1

        # return
        return number_of_rows, number_of_cols

    def _determine_table_bounding_box(
        self, lines_in_table: typing.List[LineSegment]
    ) -> Rectangle:

        # determine bounding box
        min_x: Decimal = lines_in_table[0].x0
        max_x: Decimal = lines_in_table[0].x0
        min_y: Decimal = lines_in_table[0].y0
        max_y: Decimal = lines_in_table[0].y0
        for l in lines_in_table:
            min_x = min([min_x, l.x0, l.x1])
            max_x = max([max_x, l.x0, l.x1])
            min_y = min([min_y, l.y0, l.y1])
            max_y = max([max_y, l.y0, l.y1])

        # return
        return Rectangle(min_x, min_y, max_x - min_x, max_y - min_y)

    def _is_unbroken(self, r: Rectangle) -> bool:
        r0: Rectangle = r.shrink(Decimal(1))
        r1: Rectangle = r.grow(Decimal(1))
        for l in self._lines_per_page[self._current_page_number]:
            # one of the points of the line lies outside the outer rectangle, one of the points lies inside the inner rectangle
            if r0.contains(l.x0, l.y0) and not r1.contains(l.x1, l.y1):
                return False
            if r0.contains(l.x1, l.y1) and not r1.contains(l.x0, l.y0):
                return False

            # both points of the line lie outside the inner rectangle
            if not r0.contains(l.x0, l.y0) and not r0.contains(l.x1, l.y1):
                # fmt: off
                if abs(l.x0 - l.x1) <= Decimal(1)                                                           \
                        and r0.get_x() <= l.x0 <= r0.get_x() + r0.get_width()                               \
                        and min(l.y0, l.y1) <= r.get_y() and max(l.y0, l.y1) >= r.get_y() + r.get_height():
                    return False
                if abs(l.y0 - l.y1) <= Decimal(1)                                                           \
                        and r0.get_y() <= l.y0 <= r0.get_y() + r0.get_height()                              \
                        and min(l.x0, l.x1) <= r.get_x() and max(l.x0, l.x1) >= r.get_x() + r.get_width():
                    return False
                # fmt: on

        # default
        return True

    def _determine_table_cell_boundaries(
        self, lines_in_table: typing.List[LineSegment]
    ) -> Table:

        # keep track of unique xs / ys (to derive number of rows/cols)
        unique_xs: typing.Set[int] = set()
        unique_ys: typing.Set[int] = set()

        for l in lines_in_table:
            unique_xs.add(int(l.x0))
            unique_xs.add(int(l.x1))
            unique_ys.add(int(l.y0))
            unique_ys.add(int(l.y1))

        # determine number of rows and cols
        number_of_rows: int = len(unique_ys) - 1
        number_of_cols: int = len(unique_xs) - 1

        # sort unique_xs and unique_ys
        xs: typing.List[Decimal] = sorted([Decimal(x) for x in unique_xs])
        ys: typing.List[Decimal] = sorted([Decimal(y) for y in unique_ys])

        # find neighbouring cells and join wherever appropriate
        ds: disjointset = disjointset()
        for i in range(0, number_of_rows):
            for j in range(0, number_of_cols):
                ds.add((i, j))

        for c in range(0, len(xs) - 1):
            for r in range(0, len(ys) - 1):

                if c + 2 < len(xs):
                    logger.debug(
                        "attempting to merge [%d %d] with its right neighbour" % (r, c)
                    )
                    merged_with_right: Rectangle = Rectangle(
                        xs[c], ys[r], xs[c + 2] - xs[c], ys[r + 1] - ys[r]
                    )
                    if self._is_unbroken(merged_with_right):
                        logger.debug(
                            "merge [%d %d] with right [%d %d]" % (c, r, c + 1, r)
                        )
                        ds.union((r, c), (r, c + 1))

                if r + 2 < len(ys):
                    logger.debug(
                        "attempting to merge [%d %d] with its top neighbour" % (r, c)
                    )
                    merged_with_bottom: Rectangle = Rectangle(
                        xs[c], ys[r], xs[c + 1] - xs[c], ys[r + 2] - ys[r]
                    )
                    if self._is_unbroken(merged_with_bottom):
                        logger.debug(
                            "merge [%d %d] with bottom [%d %d]" % (c, r, c, r + 1)
                        )
                        ds.union((r, c), (r + 1, c))

        # extract clusters
        cells: typing.Dict[
            typing.Tuple[int, int], typing.List[typing.Tuple[int, int]]
        ] = {}
        for i in range(0, number_of_rows):
            for j in range(0, number_of_cols):
                p: typing.Tuple[int, int] = ds.find((i, j))
                if p not in cells:
                    cells[p] = []
                cells[p].append((i, j))

        # build Table
        table: Table = FlexibleColumnWidthTable(
            number_of_rows=number_of_rows, number_of_columns=number_of_cols
        )
        for _, v in cells.items():
            min_row: int = min([int_tuple[0] for int_tuple in v])
            max_row: int = max([int_tuple[0] for int_tuple in v])
            min_col: int = min([int_tuple[1] for int_tuple in v])
            max_col: int = max([int_tuple[1] for int_tuple in v])

            # check whether all areas are rectangular
            for i in range(min_col, max_col):
                for j in range(min_row, max_row):
                    assert (
                        j * number_of_rows + i
                    ) in v, "Non-rectangular area detected in table."

            # create TableCell
            tc: TableCell = TableCell(
                Paragraph(" "),
                row_span=max_row - min_row + 1,
                col_span=max_col - min_col + 1,
            )

            # set bounding_box
            tc._previous_layout_box = Rectangle(
                xs[min_col],
                ys[min_row],
                xs[max_col + 1] - xs[min_col],
                ys[max_row + 1] - ys[min_row],
            )
            tc._previous_paint_box = tc._previous_layout_box

            # add to Table
            table.add(tc)

        # return
        return table



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
    The cross-reference table contains information that permits random access to indirect objects within the file so
    that the entire file need not be read to locate any particular object.
"""
import io
from typing import List, Optional, Union

from borb.io.read.tokenize.high_level_tokenizer import HighLevelTokenizer
from borb.io.read.tokenize.low_level_tokenizer import TokenType
from borb.io.read.types import Dictionary, Name, Reference
from borb.pdf.xref.xref import XREF


class PlainTextXREF(XREF):
    """
    The cross-reference table contains information that permits random access to indirect objects within the file so
    that the entire file need not be read to locate any particular object. The table shall contain a one-line entry for
    each indirect object, specifying the byte offset of that object within the body of the file. (Beginning with PDF 1.5,
    some or all of the cross-reference information may alternatively be contained in cross-reference streams; see
    7.5.8, "Cross-Reference Streams.")

    NOTE 1
    The cross-reference table is the only part of a PDF file with a fixed format, which permits entries in the table to
    be accessed randomly.
    The table comprises one or more cross-reference sections. Initially, the entire table consists of a single section
    (or two sections if the file is linearized; see Annex F). One additional section shall be added each time the file is
    incrementally updated (see 7.5.6, "Incremental Updates").

    Each cross-reference section shall begin with a line containing the keyword xref. Following this line shall be
    one or more cross-reference subsections, which may appear in any order. For a file that has never been
    incrementally updated, the cross-reference section shall contain only one subsection, whose object numbering
    begins at 0.
    """

    def __init__(self):
        super().__init__()

    def read(
        self,
        src: Union[io.BufferedIOBase, io.RawIOBase, io.BytesIO],
        tok: HighLevelTokenizer,
        initial_offset: Optional[int] = None,
    ) -> "XREF":
        """
        This method attempts to read a plaintext XREF from the given io_source.
        It will either throw an exception, or return this XREF
        """

        if initial_offset is not None:
            src.seek(initial_offset)
        else:
            self._seek_to_xref_token(src, tok)

        # now we should be back to the start of XREF
        token = tok.next_non_comment_token()
        assert token is not None
        assert token.get_text() == "xref"

        # read xref sections
        while True:
            xref_section = self._read_section(src, tok)
            if len(xref_section) == 0:
                break
            else:
                for r in xref_section:
                    self.add(r)

        # process trailer
        self[Name("Trailer")] = self._read_trailer(src, tok)
        self[Name("Trailer")].set_parent(self)

        # return self
        return self

    def _read_section(
        self,
        src: Union[io.BufferedIOBase, io.RawIOBase, io.BytesIO],
        tok: HighLevelTokenizer,
    ) -> List[Reference]:

        tokens = [tok.next_non_comment_token() for _ in range(0, 2)]
        assert tokens[0] is not None
        assert tokens[1] is not None
        if tokens[0].get_text() in ["trailer", "startxref"]:
            src.seek(tokens[0].get_byte_offset())
            return []
        assert tokens[0].get_token_type() == TokenType.NUMBER
        assert tokens[1].get_token_type() == TokenType.NUMBER

        start_object_number = int(tokens[0].get_text())
        number_of_objects = int(tokens[1].get_text())
        indirect_references = []

        # read subsection
        for i in range(0, number_of_objects):
            tokens = [tok.next_non_comment_token() for _ in range(0, 3)]
            assert tokens[0] is not None
            assert tokens[0].get_text() not in ["trailer", "startxref"]
            assert tokens[0].get_token_type() == TokenType.NUMBER

            assert tokens[1] is not None
            assert tokens[1].get_token_type() == TokenType.NUMBER

            assert tokens[2] is not None
            assert tokens[2].get_token_type() == TokenType.OTHER
            assert tokens[2].get_text() in ["f", "n"]

            indirect_references.append(
                Reference(
                    object_number=start_object_number + i,
                    byte_offset=int(tokens[0].get_text()),
                    generation_number=int(tokens[1].get_text()),
                    is_in_use=(tokens[2].get_text() == "n"),
                )
            )

        # return
        return indirect_references

    def _read_trailer(
        self,
        src: Union[io.BufferedIOBase, io.RawIOBase, io.BytesIO],
        tok: HighLevelTokenizer,
    ) -> Dictionary:

        # return None if there is no trailer
        token = tok.next_non_comment_token()
        assert token is not None
        if token.get_text() != "trailer":
            return Dictionary()

        # if there is a keyword "trailer" the next token should be TokenType.START_DICT
        token = tok.next_non_comment_token()
        assert token is not None
        assert token.get_token_type() == TokenType.START_DICT

        # go back 2 chars "<<"
        src.seek(-2, io.SEEK_CUR)

        # read dictionary as trailer
        trailer_dict = tok.read_dictionary()

        # process startxref
        token = tok.next_non_comment_token()
        assert token is not None
        assert token.get_token_type() == TokenType.OTHER
        assert token.get_text() == "startxref"

        # return
        return trailer_dict



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This file contains all the classes needed to perform layout of text-elements.
This includes; ChunkOfText, LineOfText, Paragraph and Heading
"""
import typing
from decimal import Decimal
from typing import Union

from borb.pdf.canvas.color.color import Color, HexColor
from borb.pdf.canvas.font.font import Font
from borb.pdf.canvas.font.glyph_line import GlyphLine
from borb.pdf.canvas.geometry.rectangle import Rectangle
from borb.pdf.canvas.layout.hyphenation.hyphenation import Hyphenation
from borb.pdf.canvas.layout.layout_element import Alignment
from borb.pdf.canvas.layout.text.line_of_text import LineOfText
from borb.pdf.page.page import Page


class Paragraph(LineOfText):
    """
    A paragraph (from the Ancient Greek παράγραφος, parágraphos, "to write beside")
    is a self-contained unit of discourse in writing dealing with a particular point or idea.
    A paragraph consists of one or more sentences. Though not required by the syntax of any language,
    paragraphs are usually an expected part of formal writing, used to organize longer prose.
    """

    def __init__(
        self,
        text: str,
        respect_newlines_in_text: bool = False,
        respect_spaces_in_text: bool = False,
        font: Union[Font, str] = "Helvetica",
        font_size: Decimal = Decimal(12),
        text_alignment: Alignment = Alignment.LEFT,
        vertical_alignment: Alignment = Alignment.TOP,
        horizontal_alignment: Alignment = Alignment.LEFT,
        font_color: Color = HexColor("000000"),
        border_top: bool = False,
        border_right: bool = False,
        border_bottom: bool = False,
        border_left: bool = False,
        border_radius_top_left: Decimal = Decimal(0),
        border_radius_top_right: Decimal = Decimal(0),
        border_radius_bottom_right: Decimal = Decimal(0),
        border_radius_bottom_left: Decimal = Decimal(0),
        border_color: Color = HexColor("000000"),
        border_width: Decimal = Decimal(1),
        padding_top: Decimal = Decimal(0),
        padding_right: Decimal = Decimal(0),
        padding_bottom: Decimal = Decimal(0),
        padding_left: Decimal = Decimal(0),
        margin_top: typing.Optional[Decimal] = None,
        margin_right: typing.Optional[Decimal] = None,
        margin_bottom: typing.Optional[Decimal] = None,
        margin_left: typing.Optional[Decimal] = None,
        fixed_leading: typing.Optional[Decimal] = None,
        multiplied_leading: typing.Optional[Decimal] = None,
        background_color: typing.Optional[Color] = None,
        hyphenation: typing.Optional[Hyphenation] = None,
    ):
        super().__init__(
            text=text,
            font=font,
            font_size=font_size,
            vertical_alignment=vertical_alignment,
            horizontal_alignment=horizontal_alignment,
            font_color=font_color,
            border_top=border_top,
            border_right=border_right,
            border_bottom=border_bottom,
            border_left=border_left,
            border_radius_top_left=border_radius_top_left,
            border_radius_top_right=border_radius_top_right,
            border_radius_bottom_right=border_radius_bottom_right,
            border_radius_bottom_left=border_radius_bottom_left,
            border_color=border_color,
            border_width=border_width,
            padding_top=padding_top,
            padding_right=padding_right,
            padding_bottom=padding_bottom,
            padding_left=padding_left,
            margin_top=margin_top,
            margin_right=margin_right,
            margin_bottom=margin_bottom,
            margin_left=margin_left,
            multiplied_leading=multiplied_leading,
            fixed_leading=fixed_leading,
            background_color=background_color,
        )
        self._respect_newlines_in_text = respect_newlines_in_text
        self._respect_spaces_in_text = respect_spaces_in_text
        self._hyphenation = hyphenation

        # alignment
        assert text_alignment in [
            Alignment.LEFT,
            Alignment.CENTERED,
            Alignment.RIGHT,
            Alignment.JUSTIFIED,
        ]
        self._text_alignment = text_alignment

        # layout
        self._previous_content_box: typing.Optional[Rectangle] = None
        self._previous_attr_hash_for_layout: typing.Optional[int] = None
        self._previous_lines_of_text: typing.Optional[typing.List[LineOfText]] = None

    def _split_text(self, bounding_box: Rectangle) -> typing.List[str]:
        # asserts
        assert self._font_size is not None

        # attempt to split into words (preserve space if needed)
        words: typing.List[str] = [""]
        tokens_to_split_on: typing.List[str] = [" ", "\t", "\n"]

        tokens_to_preserve: typing.List[str] = []
        if self._respect_newlines_in_text:
            tokens_to_preserve.append("\n")
        if self._respect_spaces_in_text:
            tokens_to_preserve.append(" ")
            tokens_to_preserve.append("\t")

        for c in self._text:
            if c in tokens_to_split_on:
                # we have a token we split on, and preserve
                # add it to the list of words
                if c in tokens_to_preserve:
                    words.append(c)
                    words.append("")
                else:
                    # we have a token we split on, but don't preserve
                    # such as whitespace, with self.respect_spaces_in_text set to False
                    if words[-1] != "":
                        words.append("")
            else:
                # build the word that was already being built
                words[-1] += c
        words = [x for x in words if len(x) > 0]

        # build lines using words
        lines_of_text = []
        for i, w in enumerate(words):

            # split on \n
            if w == "\n" and self._respect_newlines_in_text:
                lines_of_text.append("")
                continue

            # build line of text to check if it fits the bounding box
            potential_text = lines_of_text[-1] if len(lines_of_text) > 0 else ""
            if len(potential_text) != 0 and not self._respect_spaces_in_text:
                potential_text += " "
            potential_text += w

            # check the remaining space in the box
            # checking with 0 is not a great idea due to rounding errors
            # so, as a pre-emptive measure, we round the number to 2 digits
            # fmt: off
            potential_width = GlyphLine.from_str(potential_text, self._font, self._font_size).get_width_in_text_space()
            remaining_space_in_box: Decimal = round(bounding_box.width - potential_width, 2)
            # fmt: on

            # IF there is space left over, we add the word to the lines of text being built
            if remaining_space_in_box >= Decimal(0):
                if len(lines_of_text) == 0:
                    lines_of_text.append(w)
                else:
                    if len(lines_of_text[-1]) > 0 and not self._respect_spaces_in_text:
                        lines_of_text[-1] += " "
                    lines_of_text[-1] += w

            # (ELSE) there is no more room in the box for this word,
            # BUT perhaps we can hyphenate the word
            else:
                # if no hyphenation class is provided, we can't hyphenate
                if self._hyphenation is None:
                    lines_of_text.append(w)
                    continue

                # if we have to respect the spacing in the text, we don't hyphenate
                if self._respect_spaces_in_text:
                    lines_of_text.append(w)
                    continue

                # calculate potential hyphenation breaks
                # if the word can not be broken into parts, we can't hyphenate
                hyphenated_word_parts = self._hyphenation.hyphenate(w).split(chr(173))
                if len(hyphenated_word_parts) == 1:
                    lines_of_text.append(w)
                    continue

                potential_text = lines_of_text[-1] if len(lines_of_text) > 0 else ""
                if len(potential_text) != 0 and not self._respect_spaces_in_text:
                    potential_text += " "

                # check where the text can be split, in order to fit in the bounding box
                hyphenation_split_index: int = 0
                for i in range(1, len(hyphenated_word_parts)):
                    # fmt: off
                    potential_text_after_hyphenation = potential_text + "".join([x for x in hyphenated_word_parts[0:i]]) + "-"
                    potential_width = GlyphLine.from_str(potential_text_after_hyphenation, self._font, self._font_size).get_width_in_text_space()
                    remaining_space_in_box = round(bounding_box.width - potential_width, 2)
                    # fmt: on
                    if remaining_space_in_box > Decimal(0):
                        hyphenation_split_index = i
                    else:
                        break

                # no sensible split was found
                if hyphenation_split_index == 0:
                    lines_of_text.append(w)
                    continue

                # break the text according to the hyphenation
                # fmt: off
                if len(lines_of_text[-1]) > 0 and not self._respect_spaces_in_text:
                    lines_of_text[-1] += " "
                lines_of_text[-1] += "".join([x for x in hyphenated_word_parts[0:hyphenation_split_index]]) + "-"
                lines_of_text.append("".join([x for x in hyphenated_word_parts[hyphenation_split_index:]]))
                # fmt: on

        # last-minute cleanup
        while len(lines_of_text) > 0 and lines_of_text[-1] == "":
            lines_of_text.pop(len(lines_of_text) - 1)

        # return
        return lines_of_text if len(lines_of_text) > 0 else [""]

    #
    #  RENDERING LOGIC
    #

    def _calculate_attribute_hash_for_content_box(self) -> int:
        attr: typing.List[typing.Any] = [
            self._text,
            self._font,
            self._font_size,
            self._vertical_alignment,
            self._horizontal_alignment,
            self._font_color,
            self._border_top,
            self._border_right,
            self._border_bottom,
            self._border_left,
            self._border_radius_top_left,
            self._border_radius_top_right,
            self._border_radius_bottom_right,
            self._border_radius_bottom_left,
            self._border_color,
            self._border_width,
            self._padding_top,
            self._padding_right,
            self._padding_bottom,
            self._padding_left,
            self._margin_top,
            self._margin_right,
            self._margin_bottom,
            self._margin_left,
            self._multiplied_leading,
            self._fixed_leading,
            self._background_color,
        ]
        attr_hash_for_layout: int = 1927868237
        for a in attr:
            h: int = hash(a)
            attr_hash_for_layout ^= (h ^ (h << 16) ^ 89869747) * 3644798167
        attr_hash_for_layout = attr_hash_for_layout * 69069 + 907133923

        # return
        return attr_hash_for_layout

    def _get_content_box(self, available_space: Rectangle) -> Rectangle:

        # compare hash
        attr_hash_for_layout: int = self._calculate_attribute_hash_for_content_box()

        # check whether we are calculating the available space
        # using the same (Paragraph) attributes using a previous content box
        # if so, return the previous content_box
        # if this Paragraph previously fit in that box, and nothing has changed (e.g. font_size, leading, etc)
        # then this Paragraph still fits inside that box
        if (
            self._previous_content_box is not None
            and available_space == self._previous_content_box
            and attr_hash_for_layout == self._previous_attr_hash_for_layout
        ):
            return self._previous_content_box

        # if the hash has changed, it implies some attributes that are relevant to layout have changed
        # we need to recalculate the content_box
        # otherwise we can just skip this and return the previously calculated content_box
        assert self._font_size is not None
        self._previous_attr_hash_for_layout = attr_hash_for_layout
        self._previous_lines_of_text = [
            LineOfText(
                x,
                font=self._font,
                font_size=self._font_size,
                font_color=self._font_color,
                horizontal_alignment=self._text_alignment,
                multiplied_leading=self._multiplied_leading,
                text_alignment=self._text_alignment,
                fixed_leading=self._fixed_leading,
            )
            for x in self._split_text(available_space)
        ]

        #  When using justification,
        #  it is customary to treat the last line of a paragraph separately by simply left or right aligning it,
        #  depending on the language direction.
        if (
            self._text_alignment == Alignment.JUSTIFIED
            and len(self._previous_lines_of_text) > 1
        ):
            self._previous_lines_of_text[-1]._text_alignment = Alignment.LEFT

        # determine line height
        assert self._font_size is not None
        line_height: Decimal = self._font_size
        if self._multiplied_leading is not None:
            line_height *= self._multiplied_leading
        if self._fixed_leading is not None:
            line_height += self._fixed_leading

        # determine content box height
        h: Decimal = line_height * len(self._previous_lines_of_text)

        # determine content box width
        w: Decimal = Decimal(0)
        for i, l in enumerate(self._previous_lines_of_text):
            lbox: Rectangle = l._get_content_box(
                Rectangle(
                    available_space.get_x(),
                    available_space.get_y()
                    + available_space.get_height()
                    - (i + 1) * line_height,
                    available_space.get_width(),
                    line_height,
                )
            )
            w = max(w, lbox.get_width())

        # determine content box
        self._previous_content_box = Rectangle(
            available_space.get_x(),
            available_space.get_y() + available_space.get_height() - h,
            w,
            h,
        )

        # return
        return self._previous_content_box

    def _paint_content_box(self, page: Page, available_space: Rectangle) -> None:

        # ensure everything is initialized
        self._get_content_box(available_space)

        # determine line height
        assert self._font_size is not None
        line_height: Decimal = self._font_size
        if self._multiplied_leading is not None:
            line_height *= self._multiplied_leading
        if self._fixed_leading is not None:
            line_height += self._fixed_leading

        # call paint on all LineOfText objects
        for i, l in enumerate(self._previous_lines_of_text):
            l.paint(
                page,
                Rectangle(
                    available_space.get_x(),
                    available_space.get_y()
                    + available_space.get_height()
                    - (i + 1) * line_height,
                    available_space.get_width(),
                    line_height,
                ),
            )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This implementation of LayoutElement represents a Chart
"""
import io
import typing
from decimal import Decimal
from typing import Optional

from PIL import Image as PILImage  # type: ignore [import]

from borb.pdf.canvas.color.color import HexColor, Color
from borb.pdf.canvas.layout.image.image import Image
from borb.pdf.canvas.layout.layout_element import Alignment


class Chart(Image):
    """
    This implementation of LayoutElement represents a Chart
    """

    def __init__(
        self,
        chart: "matplotlib.pyplot",  # type: ignore[name-defined]
        border_bottom: bool = False,
        border_color: Color = HexColor("000000"),
        border_left: bool = False,
        border_radius_bottom_left: Decimal = Decimal(0),
        border_radius_bottom_right: Decimal = Decimal(0),
        border_radius_top_left: Decimal = Decimal(0),
        border_radius_top_right: Decimal = Decimal(0),
        border_right: bool = False,
        border_top: bool = False,
        border_width: Decimal = Decimal(1),
        height: Optional[Decimal] = None,
        horizontal_alignment: Alignment = Alignment.LEFT,
        margin_bottom: typing.Optional[Decimal] = None,
        margin_left: typing.Optional[Decimal] = None,
        margin_right: typing.Optional[Decimal] = None,
        margin_top: typing.Optional[Decimal] = None,
        padding_bottom: Decimal = Decimal(0),
        padding_left: Decimal = Decimal(0),
        padding_right: Decimal = Decimal(0),
        padding_top: Decimal = Decimal(0),
        vertical_alignment: Alignment = Alignment.TOP,
        width: Optional[Decimal] = None,
    ):
        # chart to image
        byte_buffer = io.BytesIO()
        chart.savefig(byte_buffer, format="png")
        byte_buffer.seek(0)

        super(Chart, self).__init__(
            image=PILImage.open(byte_buffer),
            width=width,
            height=height,
            border_bottom=border_bottom,
            border_color=border_color,
            border_left=border_left,
            border_radius_bottom_left=border_radius_bottom_left,
            border_radius_bottom_right=border_radius_bottom_right,
            border_radius_top_left=border_radius_top_left,
            border_radius_top_right=border_radius_top_right,
            border_right=border_right,
            border_top=border_top,
            border_width=border_width,
            horizontal_alignment=horizontal_alignment,
            margin_bottom=margin_bottom if margin_bottom is not None else Decimal(5),
            margin_left=margin_left if margin_left is not None else Decimal(5),
            margin_right=margin_right if margin_right is not None else Decimal(5),
            margin_top=margin_top if margin_top is not None else Decimal(5),
            padding_bottom=padding_bottom,
            padding_left=padding_left,
            padding_right=padding_right,
            padding_top=padding_top,
            vertical_alignment=vertical_alignment,
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
A maze where every point is reachable and where there is only one single path
from one point in the maze to any other point is called a perfect maze.
"""
import random
import typing
from decimal import Decimal


class RectangularMazeFactory:
    """
    A maze where every point is reachable and where there is only one single path
    from one point in the maze to any other point is called a perfect maze.
    """

    @staticmethod
    def rectangular_maze(grid_width: int = 10, grid_height: int = 10):
        """
        This function generates a perfect rectangular maze.
        :param grid_width:      the width of the grid (expressed in cells)
        :param grid_height:     the height of the grid (expressed in cells)
        :return:                a perfect maze, represented as a typing.List[typing.Tuple[typing.Tuple[Decimal, Decimal], typing.Tuple[Decimal, Decimal]]]
        """

        # determine lines
        cells: typing.List[typing.List[typing.Tuple[bool, bool, bool, bool]]] = [
            [(True, True, True, True) for _ in range(0, grid_width)]
            for _ in range(0, grid_height)
        ]

        cells_stk: typing.List[typing.Tuple[int, int]] = [
            (
                random.randint(0, grid_width - 1),
                random.randint(0, grid_height - 1),
            )
        ]
        while sum([sum([1 for t in row if all(t)]) for row in cells]) > 0:
            # check last element
            x, y = cells_stk[-1]

            # find intact neighbours
            nbs: typing.List[typing.Tuple[int, int]] = []
            if x != 0 and all(cells[x - 1][y]):
                nbs.append((x - 1, y))
            if x != (grid_width - 1) and all(cells[x + 1][y]):
                nbs.append((x + 1, y))
            if y != 0 and all(cells[x][y - 1]):
                nbs.append((x, y - 1))
            if y != (grid_height - 1) and all(cells[x][y + 1]):
                nbs.append((x, y + 1))
            # pop
            if len(nbs) == 0:
                cells_stk.pop(-1)
                continue
            # go to unvisited nb
            nb: typing.Tuple[int, int] = random.choice(nbs)
            cells_stk.append(nb)

            # tear down the walls
            # new cell is WEST
            if x == nb[0] + 1 and y == nb[1]:
                cells[x - 1][y] = (
                    cells[x - 1][y][0],
                    False,
                    cells[x - 1][y][2],
                    cells[x - 1][y][3],
                )
                cells[x][y] = (cells[x][y][0], cells[x][y][1], cells[x][y][2], False)
            # new cell is EAST
            if x == nb[0] - 1 and y == nb[1]:
                cells[x + 1][y] = (
                    cells[x + 1][y][0],
                    cells[x + 1][y][1],
                    cells[x + 1][y][2],
                    False,
                )
                cells[x][y] = (cells[x][y][0], False, cells[x][y][2], cells[x][y][3])
            # new cell is NORTH
            if x == nb[0] and y == nb[1] + 1:
                cells[x][y - 1] = (
                    cells[x][y - 1][0],
                    cells[x][y - 1][0],
                    False,
                    cells[x][y - 1][3],
                )
                cells[x][y] = (False, cells[x][y][1], cells[x][y][2], cells[x][y][3])
            # new cell is SOUTH
            if x == nb[0] and y == nb[1] - 1:
                cells[x][y + 1] = (
                    False,
                    cells[x][y + 1][1],
                    cells[x][y + 1][2],
                    cells[x][y + 1][3],
                )
                cells[x][y] = (cells[x][y][0], cells[x][y][1], False, cells[x][y][3])
        lines: typing.List[
            typing.Tuple[typing.Tuple[Decimal, Decimal], typing.Tuple[Decimal, Decimal]]
        ] = []
        for i in range(0, grid_width):
            for j in range(0, grid_height):
                # NORTH
                if cells[i][j][0]:
                    lines.append(
                        (
                            (Decimal(i * 10), Decimal(j * 10)),
                            (Decimal((i + 1) * 10), Decimal(j * 10)),
                        )
                    )
                # EAST
                if cells[i][j][1]:
                    lines.append(
                        (
                            (Decimal((i + 1) * 10), Decimal(j * 10)),
                            (Decimal((i + 1) * 10), Decimal((j + 1) * 10)),
                        )
                    )
                # SOUTH
                if cells[i][j][2]:
                    lines.append(
                        (
                            (Decimal((i + 1) * 10), Decimal((j + 1) * 10)),
                            (Decimal(i * 10), Decimal((j + 1) * 10)),
                        )
                    )
                # WEST
                if cells[i][j][3]:
                    lines.append(
                        (
                            (Decimal(i * 10), Decimal((j + 1) * 10)),
                            (Decimal(i * 10), Decimal(j * 10)),
                        )
                    )

        # return
        return lines



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
A composite font, also called a Type 0 font, is one whose glyphs are obtained from a fontlike object called a
CIDFont. A composite font shall be represented by a font dictionary whose Subtype value is Type0. The Type
0 font is known as the root font, and its associated CIDFont is called its descendant.
"""
import logging
import typing
from pathlib import Path

from borb.io.read.types import Decimal as bDecimal
from borb.io.read.types import List, Name, Stream
from borb.pdf.canvas.font.font import Font

logger = logging.getLogger(__name__)


class Type0Font(Font):
    """
    A composite font, also called a Type 0 font, is one whose glyphs are obtained from a fontlike object called a
    CIDFont. A composite font shall be represented by a font dictionary whose Subtype value is Type0. The Type
    0 font is known as the root font, and its associated CIDFont is called its descendant.
    """

    def __init__(self):
        super(Type0Font, self).__init__()
        self[Name("Type")] = Name("Font")
        self[Name("Subtype")] = Name("Type0")
        self._character_identifier_to_unicode_lookup: typing.Dict[int, str] = {}
        self._unicode_lookup_to_character_identifier: typing.Dict[str, int] = {}
        self._byte_to_char_identifier: typing.Dict[int, int] = {}

    def _read_to_unicode(self):
        if len(self._unicode_lookup_to_character_identifier) > 0:
            return
        # fmt: off
        assert "ToUnicode" in self, "Type0Font must have a /ToUnicode entry"
        assert ("DecodedBytes" in self["ToUnicode"]), "Type0Font must have a valid /ToUnicode entry"
        # fmt: on
        cmap_bytes: bytes = self["ToUnicode"]["DecodedBytes"]
        self._character_identifier_to_unicode_lookup = self._read_cmap(cmap_bytes)
        self._unicode_lookup_to_character_identifier: typing.Dict[str, int] = {}
        for k, v in self._character_identifier_to_unicode_lookup.items():
            if v not in self._unicode_lookup_to_character_identifier:
                self._unicode_lookup_to_character_identifier[v] = k

    def _read_encoding_cmap(self):
        if len(self._byte_to_char_identifier) > 0:
            return
        # fmt: off
        assert "Encoding" in self, "Type0Font must have an /Encoding entry"
        assert ("DecodedBytes" in self["Encoding"]), "Type0Font must have a valid /Encoding entry"
        # fmt: on
        cmap_bytes: bytes = self["Encoding"]["DecodedBytes"]
        self._byte_to_char_identifier = {
            k: v for k, v in self._read_cmap(cmap_bytes).items()
        }
        self._char_to_byte_identifier = {
            v: k for k, v in self._byte_to_char_identifier.items()
        }

    def character_identifier_to_unicode(
        self, character_identifier: int
    ) -> typing.Optional[str]:
        """
        This function maps a character identifier to its unicode str.
        If no such mapping exists, this function returns None.
        """

        # If the font dictionary contains a ToUnicode CMap (see 9.10.3, "ToUnicode CMaps"), use that CMap to
        # convert the character code to Unicode.
        if Name("ToUnicode") in self:
            self._read_to_unicode()
            return self._character_identifier_to_unicode_lookup.get(
                character_identifier
            )

        if Name("Encoding") in self:

            # a) Map the character code to a character identifier (CID) according to the font’s CMap.
            cid: typing.Optional[int] = None
            if isinstance(self["Encoding"], Name):
                encoding_name: str = str(self["Encoding"])
                assert encoding_name in ["Identity", "Identity-H"]
                cid = character_identifier
            if isinstance(self["Encoding"], Stream):
                self._read_encoding_cmap()
                cid = self._byte_to_char_identifier.get(character_identifier)
            if cid is None:
                return None
            assert cid is not None

            # d) Obtain the CMap with the name constructed in step (c) (available from the ASN Web site; see the
            # Bibliography).
            # https://github.com/adobe-type-tools/cmap-resources
            if len(self._character_identifier_to_unicode_lookup) == 0:
                self._character_identifier_to_unicode_lookup = (
                    Type0Font._find_best_matching_predefined_cmap(self._get_cmap_name())
                )
                self._unicode_lookup_to_character_identifier = {
                    v: k
                    for k, v in self._character_identifier_to_unicode_lookup.items()
                }

            # e) Map the CID obtained in step (a) according to the CMap obtained in step (d), producing a
            # Unicode value.
            return self._character_identifier_to_unicode_lookup.get(cid, None)

        # default
        return None

    def _get_cmap_name(self) -> str:
        # b) Obtain the registry and ordering of the character collection used by the font’s CMap
        # (for example, Adobe and Japan1) from its CIDSystemInfo dictionary.
        # fmt: off
        assert "DescendantFonts" in self, "Type0Font must have a /DescendantFonts entry"
        assert isinstance(self["DescendantFonts"], List), "Type0Font must have a valid /DescendantFonts entry"
        assert (len(self["DescendantFonts"]) == 1), "Type0Font must have a valid /DescendantFonts entry"
        assert ("CIDSystemInfo" in self["DescendantFonts"][0]), "Type0Font must have a valid /DescendantFonts entry"
        assert ("Registry" in self["DescendantFonts"][0]["CIDSystemInfo"]), "Type0Font must have a valid /DescendantFonts entry"
        assert ("Ordering" in self["DescendantFonts"][0]["CIDSystemInfo"]), "Type0Font must have a valid /DescendantFonts entry"
        # fmt: on
        registry: str = str(self["DescendantFonts"][0]["CIDSystemInfo"]["Registry"])
        ordering: str = str(self["DescendantFonts"][0]["CIDSystemInfo"]["Ordering"])

        # c) Construct a second CMap name by concatenating the registry and ordering obtained in step (b) in
        # the format registry–ordering–UCS2 (for example, Adobe–Japan1–UCS2).
        cmap_name: str = "".join([registry, "-", ordering, "-", "UCS2"])

        return cmap_name

    @staticmethod
    def _find_best_matching_predefined_cmap(cmap_name: str) -> typing.Dict[int, str]:
        cmap_dir: Path = Path(__file__).parent / "cmaps"
        assert cmap_dir.exists(), "cmaps dir not found."
        predefined_cmaps: typing.List[str] = [x.name for x in cmap_dir.iterdir()]

        # pseudo match
        if cmap_name not in predefined_cmaps:

            if cmap_name == "Adobe-Identity-UCS2":
                # fmt: off
                logger.info("Encoding Adobe-Identity-UCS2 was specified, using Adobe-Identity-H in stead")
                cmap_name = "Adobe-Identity-H"
                # fmt: on

            if cmap_name == "Adobe-Japan1-UCS2":
                # fmt: off
                logger.info("Encoding Adobe-Identity-UCS2 was specified, using Adobe-Japan1-0 in stead")
                cmap_name = "Adobe-Japan1-0"
                # fmt: on

            if cmap_name not in predefined_cmaps:
                # fmt: off
                logger.info("Encoding %s was specified, defaulting to Adobe-Identity-H in stead" % cmap_name)
                cmap_name = "Adobe-Identity-H"
                # fmt: on

        # read predefined cmap
        cmap_bytes: typing.Optional[bytes] = None
        with open(cmap_dir / cmap_name, "rb") as cmap_file_handle:
            cmap_bytes = cmap_file_handle.read()
        assert cmap_bytes is not None

        # use Font._read_cmap to process (and return)
        return Font._read_cmap(cmap_bytes)

    def unicode_to_character_identifier(self, unicode: str) -> typing.Optional[int]:
        """
        This function maps a unicode str to its character identifier.
        If no such mapping exists, this function returns None.
        """
        if Name("ToUnicode") in self:
            self._read_to_unicode()
            return self._unicode_lookup_to_character_identifier.get(unicode)

        if Name("Encoding") in self:
            # fmt: off
            assert str(self["Encoding"]) in ["Identity", "Identity-H"], "Only Identity and Identity-H are currently supported."
            # fmt: on
            if len(self._character_identifier_to_unicode_lookup) == 0:
                self._character_identifier_to_unicode_lookup = (
                    Type0Font._find_best_matching_predefined_cmap(self._get_cmap_name())
                )
                self._unicode_lookup_to_character_identifier = {
                    v: k
                    for k, v in self._character_identifier_to_unicode_lookup.items()
                }

            # e) Map the CID obtained in step (a) according to the CMap obtained in step (d), producing a
            # Unicode value.
            return self._unicode_lookup_to_character_identifier.get(unicode, None)

        # default
        return None

    def get_width(self, character_identifier: int) -> typing.Optional[bDecimal]:
        """
        This function returns the width (in text space) of a given character identifier.
        If this Font is unable to represent the glyph that corresponds to the character identifier,
        this function returns None
        """
        # fmt: off
        assert "DescendantFonts" in self, "Type0Font must have a /DescendantFonts entry"
        assert isinstance(self["DescendantFonts"], List), "Type0Font must have a valid /DescendantFonts entry"
        assert (len(self["DescendantFonts"]) == 1), "Type0Font must have a valid /DescendantFonts entry"
        # fmt: on
        descendant_font: Font = self["DescendantFonts"][0]
        return descendant_font.get_width(character_identifier)

    def get_ascent(self) -> bDecimal:
        """
        This function returns the maximum height above the baseline reached by glyphs in this font.
        The height of glyphs for accented characters shall be excluded.
        """
        # fmt: off
        assert "DescendantFonts" in self, "Type0Font must have a /DescendantFonts entry"
        assert isinstance(self["DescendantFonts"], List), "Type0Font must have a valid /DescendantFonts entry"
        assert (len(self["DescendantFonts"]) == 1), "Type0Font must have a valid /DescendantFonts entry"
        # fmt: on
        descendant_font: Font = self["DescendantFonts"][0]
        return descendant_font.get_ascent()

    def get_descent(self) -> bDecimal:
        """
        This function returns the maximum depth below the baseline reached by glyphs in this font.
        The value shall be a negative number.
        """
        # fmt: off
        assert "DescendantFonts" in self, "Type0Font must have a /DescendantFonts entry"
        assert isinstance(self["DescendantFonts"], List), "Type0Font must have a valid /DescendantFonts entry"
        assert (len(self["DescendantFonts"]) == 1), "Type0Font must have a valid /DescendantFonts entry"
        # fmt: on
        descendant_font: Font = self["DescendantFonts"][0]
        return descendant_font.get_descent()

    def _empty_copy(self) -> "Font":
        return Type0Font()

    def __deepcopy__(self, memodict={}):
        # fmt: off
        f_out: Type0Font = super(Type0Font, self).__deepcopy__(memodict)
        f_out[Name("Subtype")] = Name("Type0")
        f_out._character_identifier_to_unicode_lookup: typing.Dict[int, str] = {k: v for k, v in self._character_identifier_to_unicode_lookup.items()}
        f_out._unicode_lookup_to_character_identifier: typing.Dict[str, int] = {k: v for k, v in self._unicode_lookup_to_character_identifier.items()}
        return f_out
        # fmt: on



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This class represents a common base for all LayoutElement implementations
that attempt to represent tabular data.
"""
import copy
import typing
from decimal import Decimal

from borb.pdf.canvas.color.color import Color, HexColor
from borb.pdf.canvas.geometry.rectangle import Rectangle
from borb.pdf.canvas.layout.layout_element import Alignment, LayoutElement
from borb.pdf.page.page import Page


class TableCell(LayoutElement):
    """
    This class represents a single cell of a table
    """

    def __init__(
        self,
        layout_element: LayoutElement,
        row_span: int = 1,
        col_span: int = 1,
        background_color: typing.Optional[Color] = None,
        border_bottom: bool = True,
        border_color: Color = HexColor("000000"),
        border_left: bool = True,
        border_radius_bottom_left: Decimal = Decimal(0),
        border_radius_bottom_right: Decimal = Decimal(0),
        border_radius_top_left: Decimal = Decimal(0),
        border_radius_top_right: Decimal = Decimal(0),
        border_right: bool = True,
        border_top: bool = True,
        border_width: Decimal = Decimal(1),
        padding_bottom: Decimal = Decimal(0),
        padding_left: Decimal = Decimal(0),
        padding_right: Decimal = Decimal(0),
        padding_top: Decimal = Decimal(0),
        preferred_height: typing.Optional[Decimal] = None,
        preferred_width: typing.Optional[Decimal] = None,
    ):
        super(TableCell, self).__init__(
            background_color=background_color,
            border_bottom=border_bottom,
            border_color=border_color,
            border_left=border_left,
            border_radius_bottom_left=border_radius_bottom_left,
            border_radius_bottom_right=border_radius_bottom_right,
            border_radius_top_left=border_radius_top_left,
            border_radius_top_right=border_radius_top_right,
            border_right=border_right,
            border_top=border_top,
            border_width=border_width,
            font_size=Decimal(12),  # not used
            horizontal_alignment=Alignment.JUSTIFIED,  # not used
            margin_bottom=Decimal(0),  # not used
            margin_left=Decimal(0),  # not used
            margin_right=Decimal(0),  # not used
            margin_top=Decimal(0),  # not used
            padding_bottom=padding_bottom,
            padding_left=padding_left,
            padding_right=padding_right,
            padding_top=padding_top,
            vertical_alignment=Alignment.TOP,  # not used
        )
        self._layout_element = layout_element
        assert row_span >= 1
        assert col_span >= 1
        assert not isinstance(
            layout_element, TableCell
        ), "TableCell should not contain other TableCell LayoutElement(s)."
        assert not isinstance(
            layout_element, Table
        ), "TableCell should not contain Table LayoutElement(s)."

        # grid coordinates taken up by the TableCell
        self._row_span = row_span
        self._col_span = col_span
        self._table_coordinates: typing.List[typing.Tuple[int, int]] = []

        # width of the TableCell
        self._min_width: typing.Optional[Decimal] = None
        self._max_width: typing.Optional[Decimal] = None
        self._preferred_width: typing.Optional[Decimal] = preferred_width

        # height of the TableCell
        self._min_height: typing.Optional[Decimal] = None
        self._max_height: typing.Optional[Decimal] = None
        self._preferred_height: typing.Optional[Decimal] = preferred_height

        # layout
        self._forced_layout_box: typing.Optional[Rectangle] = None

    def get_layout_box(self, available_space: Rectangle):
        """
        This function returns the previous result of layout
        :return:    the Rectangle that was the result of the previous layout operation
        """
        if self._forced_layout_box is not None:
            return self._forced_layout_box
        return super(TableCell, self).get_layout_box(available_space)

    def _get_content_box(self, available_space: Rectangle) -> Rectangle:
        if self._forced_layout_box is not None:

            horizontal_border_width: Decimal = Decimal(0)
            if self._border_left:
                horizontal_border_width += self._border_width
            if self._border_right:
                horizontal_border_width += self._border_width

            vertical_border_width: Decimal = Decimal(0)
            if self._border_top:
                vertical_border_width += self._border_width
            if self._border_bottom:
                vertical_border_width += self._border_width

            return Rectangle(
                self._forced_layout_box.get_x()
                + self._padding_left
                + (self._border_width if self._border_left else Decimal(0)),
                self._forced_layout_box.get_y()
                + self._padding_bottom
                + (self._border_width if self._border_bottom else Decimal(0)),
                max(
                    Decimal(0),
                    self._forced_layout_box.get_width()
                    - self._padding_left
                    - self._padding_right
                    - horizontal_border_width,
                ),
                max(
                    Decimal(0),
                    self._forced_layout_box.get_height()
                    - self._padding_top
                    - self._padding_bottom
                    - vertical_border_width,
                ),
            )

        # default
        return self._layout_element.get_layout_box(available_space)

    def _set_layout_box(self, layout_box: Rectangle) -> "TableCell":
        self._forced_layout_box = layout_box
        return self

    def _paint_content_box(self, page: "Page", available_space: Rectangle) -> None:
        self._layout_element.paint(page, available_space)

    def _calculate_min_and_max_layout_box(self) -> None:
        lbox_max: Rectangle = self.get_layout_box(
            Rectangle(Decimal(0), Decimal(0), Decimal(2048), Decimal(2048))
        )
        upper_bound_lbox_min: Rectangle = copy.deepcopy(lbox_max)
        lower_bound_lbox_min: Rectangle = copy.deepcopy(
            Rectangle(
                lbox_max.get_x(), lbox_max.get_y(), Decimal(1), lbox_max.get_height()
            )
        )
        lbox_min: typing.Optional[Rectangle] = None
        while (
            abs(upper_bound_lbox_min.get_width() - lower_bound_lbox_min.get_width()) > 1
        ):
            midpoint_lbox_min: Rectangle = Rectangle(
                upper_bound_lbox_min.get_x(),
                upper_bound_lbox_min.get_y(),
                (upper_bound_lbox_min.get_width() + lower_bound_lbox_min.get_width())
                / 2,
                upper_bound_lbox_min.get_height(),
            )
            try:
                lbox_tmp: Rectangle = self.get_layout_box(midpoint_lbox_min)
                lbox_min = lbox_tmp
                if lbox_tmp.get_width() > midpoint_lbox_min.get_width():
                    lower_bound_lbox_min = midpoint_lbox_min
                else:
                    upper_bound_lbox_min = midpoint_lbox_min
            except:
                lower_bound_lbox_min = midpoint_lbox_min
                continue

        # set properties
        if lbox_min is None:
            lbox_min = lbox_max
        self._min_height = lbox_max.get_height()
        self._min_width = lbox_min.get_width()
        self._max_height = lbox_min.get_height()
        self._max_width = lbox_max.get_width()

        # return


class Table(LayoutElement):
    """
    This class represents a common base for all LayoutElement implementations
    that attempt to represent tabular data.
    """

    def __init__(
        self,
        number_of_rows: int,
        number_of_columns: int,
        background_color: typing.Optional[Color] = None,
        border_bottom: bool = False,
        border_color: Color = HexColor("000000"),
        border_left: bool = False,
        border_radius_bottom_left: Decimal = Decimal(0),
        border_radius_bottom_right: Decimal = Decimal(0),
        border_radius_top_left: Decimal = Decimal(0),
        border_radius_top_right: Decimal = Decimal(0),
        border_right: bool = False,
        border_top: bool = False,
        border_width: Decimal = Decimal(1),
        horizontal_alignment: Alignment = Alignment.LEFT,
        margin_bottom: Decimal = Decimal(0),
        margin_left: Decimal = Decimal(0),
        margin_right: Decimal = Decimal(0),
        margin_top: Decimal = Decimal(0),
        padding_bottom: Decimal = Decimal(0),
        padding_left: Decimal = Decimal(0),
        padding_right: Decimal = Decimal(0),
        padding_top: Decimal = Decimal(0),
        vertical_alignment: Alignment = Alignment.TOP,
    ):
        super(Table, self).__init__(
            background_color=background_color,
            border_bottom=border_bottom,
            border_color=border_color,
            border_left=border_left,
            border_radius_bottom_left=border_radius_bottom_left,
            border_radius_bottom_right=border_radius_bottom_right,
            border_radius_top_left=border_radius_top_left,
            border_radius_top_right=border_radius_top_right,
            border_right=border_right,
            border_top=border_top,
            border_width=border_width,
            font_size=Decimal(12),
            horizontal_alignment=horizontal_alignment,
            margin_bottom=margin_bottom,
            margin_left=margin_left,
            margin_right=margin_right,
            margin_top=margin_top,
            padding_bottom=padding_bottom,
            padding_left=padding_left,
            padding_right=padding_right,
            padding_top=padding_top,
            vertical_alignment=vertical_alignment,
        )
        assert number_of_rows >= 1
        assert number_of_columns >= 1
        self._number_of_rows = number_of_rows
        self._number_of_columns = number_of_columns
        self._content: typing.List[TableCell] = []

    def set_background_color_on_all_cells(self, background_color: Color) -> "Table":
        """
        This method sets the background Color on all TableCell objects in this Table
        """
        for e in self._content:
            e._background_color = background_color
        return self

    def set_border_width_on_all_cells(self, border_width: Decimal) -> "Table":
        """
        This method sets the border width on all TableCell objects in this Table
        """
        assert border_width >= 0
        for e in self._content:
            e._border_width = border_width
        return self

    def set_border_color_on_all_cells(self, border_color: Color) -> "Table":
        """
        This method sets the border color on all TableCell objects in this Table
        """
        for e in self._content:
            e._border_color = border_color
        return self

    def set_padding_on_all_cells(
        self,
        padding_top: Decimal,
        padding_right: Decimal,
        padding_bottom: Decimal,
        padding_left: Decimal,
    ) -> "Table":
        """
        This method sets the padding on all TableCell objects in this Table
        """
        for e in self._content:
            e._padding_top = padding_top
            e._padding_right = padding_right
            e._padding_bottom = padding_bottom
            e._padding_left = padding_left
        return self

    def set_borders_on_all_cells(
        self,
        border_top: bool,
        border_right: bool,
        border_bottom: bool,
        border_left: bool,
    ) -> "Table":
        """
        This method sets the border(s) on all TableCell objects in this Table
        """
        for e in self._content:
            e._border_top = border_top
            e._border_right = border_right
            e._border_bottom = border_bottom
            e._border_left = border_left
        return self

    def no_borders(self) -> "Table":
        """
        This method unsets the border(s) on all TableCell objects in this Table
        """
        self.set_borders_on_all_cells(False, False, False, False)
        return self

    def outer_borders(self) -> "Table":
        """
        This method unsets the border(s) on all TableCell objects in this Table
        except for the borders that form the outside edge of the Table
        """
        self.no_borders()
        for c in self._get_cells_at_row(0):
            c._border_top = True
        for c in self._get_cells_at_row(self._number_of_rows - 1):
            c._border_bottom = True
        for c in self._get_cells_at_column(0):
            c._border_left = True
        for c in self._get_cells_at_column(self._number_of_columns - 1):
            c._border_right = True
        return self

    def internal_borders(self) -> "Table":
        """
        This method sets the border(s) on all TableCell objects in this Table
        except for the borders that form the outside edge of the Table
        """
        for tc in self._content:
            tc._border_top = True
            tc._border_right = True
            tc._border_bottom = True
            tc._border_left = True
        for c in self._get_cells_at_row(0):
            c._border_top = False
        for c in self._get_cells_at_row(self._number_of_rows - 1):
            c._border_bottom = False
        for c in self._get_cells_at_column(0):
            c._border_left = False
        for c in self._get_cells_at_column(self._number_of_columns - 1):
            c._border_right = False
        return self

    def outer_borders_rounded(self, border_radius: Decimal = Decimal(20)) -> "Table":
        """
        This function sets rounded borders on the outside of this Table
        :param border_radius:   the desired border radius
        :return:                self
        """
        # upper left
        ul: typing.Optional[TableCell] = self._get_cells_at(0, 0)
        if ul is not None:
            ul._border_radius_top_left = border_radius
        # upper right
        ur: typing.Optional[TableCell] = self._get_cells_at(
            0, self._number_of_columns - 1
        )
        if ur is not None:
            ur._border_radius_top_right = border_radius
        # lower right
        lr: typing.Optional[TableCell] = self._get_cells_at(
            self._number_of_rows - 1, self._number_of_columns - 1
        )
        if lr is not None:
            lr._border_radius_bottom_right = border_radius
        # lower left
        ll: typing.Optional[TableCell] = self._get_cells_at(self._number_of_rows - 1, 0)
        if ll is not None:
            ll._border_radius_bottom_left = border_radius
        # return
        return self

    def even_odd_row_colors(
        self,
        even_row_color: Color,
        odd_row_color: Color,
        header_row_color: typing.Optional[Color] = None,
    ) -> "Table":
        """
        This function colors the Table with the classic "zebra stripes"
        e.a. one color for all even rows, and a contrasting color for the odd rows.
        :param even_row_color:      the Color to be used for even rows
        :param odd_row_color:       the Color to be used for odd rows
        :param header_row_color:    the Color to be used for the header row, if None is specified the even_row_color will be used
        This function returns self.
        """
        if header_row_color is None:
            header_row_color = even_row_color
        assert header_row_color is not None
        for r in range(0, self._number_of_rows):
            for tc in self._get_cells_at_row(r):
                if r == 0:
                    tc._background_color = header_row_color
                else:
                    if r % 2 == 0:
                        tc._background_color = even_row_color
                    else:
                        tc._background_color = odd_row_color
        return self

    def _get_cells_at(self, row: int, column: int) -> typing.Optional[TableCell]:
        for t in self._content:
            if (
                len([p for p in t._table_coordinates if p[0] == row and p[1] == column])
                > 0
            ):
                return t
        return None

    def _get_cells_at_column(self, column: int) -> typing.List[TableCell]:
        out: typing.List[TableCell] = []
        for t in self._content:
            if len([p for p in t._table_coordinates if p[1] == column]) > 0:
                out.append(t)
        return out

    def _get_cells_at_row(self, row: int) -> typing.List[TableCell]:
        out: typing.List[TableCell] = []
        for t in self._content:
            if len([p for p in t._table_coordinates if p[0] == row]) > 0:
                out.append(t)
        return out

    def add(self, layout_element: LayoutElement) -> "Table":
        """
        This function adds the given LayoutElement to this Table.
        This function returns self.
        """

        # embed LayoutElement in TableCell (if needed)
        if not isinstance(layout_element, TableCell):
            layout_element = TableCell(layout_element)

        # add content
        self._content.append(layout_element)

        # set font_size
        inner_layout_element: LayoutElement = layout_element._layout_element
        if self._font_size is None:
            self._font_size = inner_layout_element.get_font_size()

        # determine gridpoints occupied by the new TableCell
        first_incomplete_row: int = min(
            [
                x
                for x in range(0, self._number_of_rows)
                if sum([y._col_span for y in self._get_cells_at_row(x)])
                < self._number_of_columns
            ]
        )
        # check which columns are already occupied in the current row
        occupied_cols_in_row: typing.List[int] = []
        for c in self._get_cells_at_row(first_incomplete_row):
            occupied_cols_in_row.extend(
                [x[1] for x in c._table_coordinates if x[0] == first_incomplete_row]
            )
        # the first empty column is the lowest number that does not appear in occupied_cols_in_row
        first_empty_column: int = min(
            [
                x
                for x in range(0, self._number_of_columns)
                if x not in occupied_cols_in_row
            ]
        )

        # set _table_coordinates
        for i in range(0, layout_element._row_span):
            for j in range(0, layout_element._col_span):
                layout_element._table_coordinates.append(
                    (first_incomplete_row + i, first_empty_column + j)
                )

        # return
        return self



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This implementation of ReadBaseTransformer is responsible for reading the /Catalog object
"""
import io
import typing
from typing import Any, Dict, List, Optional, Union

from borb.io.read.object.dictionary_transformer import DictionaryTransformer
from borb.io.read.transformer import ReadTransformerState, Transformer
from borb.io.read.types import AnyPDFType, Decimal, Dictionary
from borb.io.read.types import List as bList
from borb.io.read.types import Name
from borb.pdf.canvas.event.event_listener import EventListener
from borb.pdf.page.page import Page


class RootDictionaryTransformer(Transformer):
    """
    This implementation of ReadBaseTransformer is responsible for reading the /Catalog object
    """

    def can_be_transformed(
        self, object: Union[io.BufferedIOBase, io.RawIOBase, io.BytesIO, AnyPDFType]
    ) -> bool:
        """
        This function returns True if the object to be converted represents a /Catalog Dictionary
        """
        return (
            isinstance(object, Dict)
            and "Type" in object
            and object["Type"] == "Catalog"
        )

    def _re_order_pages(self, root_dictionary: dict) -> None:

        # list to hold Page objects (in order)
        pages_in_order: typing.List[Page] = []

        # stack to explore Page(s) DFS
        stack_to_handle: typing.List[AnyPDFType] = []
        stack_to_handle.append(root_dictionary["Pages"])

        # DFS
        while len(stack_to_handle) > 0:
            obj = stack_to_handle.pop(0)
            if isinstance(obj, Page):
                pages_in_order.append(obj)
            # /Pages
            if (
                isinstance(obj, Dictionary)
                and "Type" in obj
                and obj["Type"] == "Pages"
                and "Kids" in obj
                and isinstance(obj["Kids"], List)
            ):
                for k in obj["Kids"]:
                    stack_to_handle.append(k)

        # change
        root_dictionary["Pages"][Name("Kids")] = bList()
        for p in pages_in_order:
            root_dictionary["Pages"]["Kids"].append(p)
        root_dictionary["Pages"][Name("Count")] = Decimal(len(pages_in_order))

    def transform(
        self,
        object_to_transform: Union[io.BufferedIOBase, io.RawIOBase, AnyPDFType],
        parent_object: Any,
        context: Optional[ReadTransformerState] = None,
        event_listeners: typing.List[EventListener] = [],
    ) -> Any:
        """
        This function reads a /Catalog Dictionary from a byte stream
        """
        # fmt: off
        assert isinstance(object_to_transform, Dictionary), "object_to_transform must be of type Dictionary"
        # fmt: on

        # convert using Dictionary transformer
        transformed_root_dictionary: Optional[Dictionary] = None
        for t in self.get_root_transformer().get_children():
            if isinstance(t, DictionaryTransformer):
                transformed_root_dictionary = t.transform(
                    object_to_transform, parent_object, context, event_listeners
                )
                break

        assert transformed_root_dictionary is not None
        assert isinstance(transformed_root_dictionary, Dictionary)

        #
        # rebuild /Pages if needed
        #
        self._re_order_pages(transformed_root_dictionary)

        # return
        return transformed_root_dictionary



# !/usr/bin/env python
# -*- coding: utf-8 -*-

"""
    This implementation of Event is triggered when an ET instruction is being processed.
"""
from borb.pdf.canvas.event.event_listener import Event


class EndTextEvent(Event):
    """
    This implementation of Event is triggered when an ET instruction is being processed.
    """

    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
A link annotation represents either a hypertext link to a destination elsewhere in the document (see 12.3.2,
“Destinations”) or an action to be performed (12.6, “Actions”). Table 173 shows the annotation dictionary
entries specific to this type of annotation.
"""
import enum
import typing
from decimal import Decimal

from borb.io.read.types import Name, List, Decimal as bDecimal, String
from borb.pdf.canvas.color.color import Color
from borb.pdf.canvas.geometry.rectangle import Rectangle
from borb.pdf.canvas.layout.annotation.annotation import Annotation


class DestinationType(enum.Enum):
    """
    This Enum represents all possible destination types (when adding a link annotation)
    """

    FIT = Name("Fit")
    FIT_B = Name("FitB")
    FIT_B_H = Name("FitBH")
    FIT_B_V = Name("FitBV")
    FIT_H = Name("FitH")
    FIT_R = Name("FitR")
    FIT_V = Name("FitV")
    X_Y_Z = Name("XYZ")


class LinkAnnotation(Annotation):
    """
    A link annotation represents either a hypertext link to a destination elsewhere in the document (see 12.3.2,
    “Destinations”) or an action to be performed (12.6, “Actions”). Table 173 shows the annotation dictionary
    entries specific to this type of annotation.
    """

    def __init__(
        self,
        bounding_box: Rectangle,
        page: Decimal,
        destination_type: DestinationType,
        color: typing.Optional[Color] = None,
        top: typing.Optional[Decimal] = None,
        right: typing.Optional[Decimal] = None,
        bottom: typing.Optional[Decimal] = None,
        left: typing.Optional[Decimal] = None,
        zoom: typing.Optional[Decimal] = None,
        highlighting_mode: typing.Optional[str] = None,
    ):
        super(LinkAnnotation, self).__init__(bounding_box=bounding_box, color=color)

        # specific for text annotations
        self[Name("Subtype")] = Name("Link")

        # (Optional; PDF 1.1) An action that shall be performed when the link
        # annotation is activated (see 12.6, “Actions”).
        # annot[Name("A")] = None

        # (Optional; not permitted if an A entry is present) A destination that shall
        # be displayed when the annotation is activated (see 12.3.2,
        # “Destinations”).
        destination = List().set_is_inline(True)  # type: ignore [attr-defined]
        destination.append(bDecimal(page))
        destination.append(destination_type.value)
        if destination_type == DestinationType.X_Y_Z:
            assert (
                left is not None
                and bottom is None
                and right is None
                and top is not None
                and zoom is not None
            )
            destination.append(bDecimal(left))
            destination.append(bDecimal(top))
            destination.append(bDecimal(zoom))
        if destination_type == DestinationType.FIT:
            assert (
                left is None
                and bottom is None
                and right is None
                and top is None
                and zoom is None
            )
        if destination_type == DestinationType.FIT_H:
            assert (
                left is None
                and bottom is None
                and right is None
                and top is not None
                and zoom is None
            )
            destination.append(bDecimal(top))
        if destination_type == DestinationType.FIT_V:
            assert (
                left is not None
                and bottom is None
                and right is None
                and top is None
                and zoom is None
            )
            destination.append(bDecimal(left))
        if destination_type == DestinationType.FIT_R:
            assert (
                left is not None
                and bottom is not None
                and right is not None
                and top is not None
                and zoom is None
            )
            destination.append(bDecimal(left))
            destination.append(bDecimal(bottom))
            destination.append(bDecimal(right))
            destination.append(bDecimal(top))
        if destination_type == DestinationType.FIT_B_H:
            assert (
                left is None
                and bottom is None
                and right is None
                and top is not None
                and zoom is None
            )
            destination.append(bDecimal(top))
        if destination_type == DestinationType.FIT_B_V:
            assert (
                left is not None
                and bottom is None
                and right is None
                and top is None
                and zoom is None
            )
            destination.append(bDecimal(left))
        self[Name("Dest")] = destination

        # (Optional; PDF 1.2) The annotation’s highlighting mode, the visual effect
        # that shall be used when the mouse button is pressed or held down
        # inside its active area:
        # N     (None) No highlighting.
        # I     (Invert) Invert the contents of the annotation rectangle.
        # O     (Outline) Invert the annotation’s border.
        # P     (Push) Display the annotation as if it were being pushed below the surface of the page.
        if highlighting_mode is not None:
            assert highlighting_mode in ["N", "I", "O", "P"]
            self[Name("H")] = String(highlighting_mode)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This implementation of WriteBaseTransformer is responsible for writing booleans
"""
from typing import Optional

from borb.io.read.types import AnyPDFType, Boolean
from borb.io.write.transformer import Transformer, WriteTransformerState


class BooleanTransformer(Transformer):
    """
    This implementation of WriteBaseTransformer is responsible for writing booleans
    """

    def can_be_transformed(self, any: AnyPDFType):
        """
        This function returns True if the object to be converted represents a Boolean object
        """
        return isinstance(any, Boolean)

    def transform(
        self,
        object_to_transform: AnyPDFType,
        context: Optional[WriteTransformerState] = None,
    ):
        """
        This method writes a Boolean to a byte stream
        """
        # fmt: off
        assert context is not None, "context must be defined to write bool objects"
        assert context.destination is not None, "context.destination must be defined to write bool objects"
        assert isinstance(object_to_transform, Boolean), "object_to_transform must be of type Boolean"
        # fmt: on

        if bool(object_to_transform):
            context.destination.write(bytes("true", "latin1"))
        else:
            context.destination.write(bytes("false", "latin1"))
